#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import mock
import pytest

from datarobot import client


@pytest.yield_fixture(scope="function", autouse=True)
def clean_client():
    client.set_client(None)
    yield
    client.set_client(None)


@pytest.fixture(scope="function", autouse=True)
def clean_env(monkeypatch):
    monkeypatch.delenv("DATAROBOT_ENDPOINT", raising=False)
    monkeypatch.delenv("DATAROBOT_API_TOKEN", raising=False)
    monkeypatch.delenv("DATAROBOT_CONFIG_FILE", raising=False)
    return monkeypatch


@pytest.yield_fixture(scope="function", autouse=True)
def clean_user_config(tmpdir):
    path = tmpdir.mkdir("userconfig").join("drconfig.yaml")
    with mock.patch("datarobot.client._get_config_dir") as m:
        m.return_value = path.dirname
        yield path


@pytest.fixture
def env_token():
    return "env_token"


@pytest.fixture
def env_endpoint():
    return "https://env.com"


@pytest.fixture
def code_config_token():
    return "code_config_token"


@pytest.fixture
def code_config_endpoint():
    return "https://code_config.com"


@pytest.fixture
def env_config_token():
    return "env_config_token"


@pytest.fixture
def env_config_endpoint():
    return "https://env_config.com"


@pytest.fixture
def env_config_vars(env_token, env_endpoint, clean_env):
    clean_env.setenv("DATAROBOT_API_TOKEN", env_token)
    clean_env.setenv("DATAROBOT_ENDPOINT", env_endpoint)


@pytest.fixture
def default_config_token():
    """
    A config token specified in the file ~/.config/datarobot/drconfig.yaml
    """
    return "user_config_token"


@pytest.fixture
def default_config_endpoint():
    """
    A config endpoint specified in the file ~/.config/datarobot/drconfig.yaml
    """
    return "https://user_config.com"


@pytest.yield_fixture
def default_config(default_config_token, default_config_endpoint, clean_user_config):
    """
    Puts a configuration file at ~/.config/datarobot/drconfig.yaml

    Cleans it up on the way out
    """
    fcontent = "endpoint: {}\ntoken: {}".format(default_config_endpoint, default_config_token)
    clean_user_config.write(fcontent)
    yield str(clean_user_config)


@pytest.fixture(scope="function")
def config_path(code_config_endpoint, code_config_token, tmpdir):
    """
    Makes sure a config exists at a path, then returns that path

    Parameters
    ----------
    code_config_endpoint
    code_config_token
    tmpdir

    Returns
    -------

    """
    fcontent = "endpoint: {}\ntoken: {}".format(code_config_endpoint, code_config_token)
    fpath = tmpdir.join("test_config.yaml")
    fpath.write(fcontent)
    return str(fpath)


@pytest.fixture(scope="function")
def env_config_path(env_config_endpoint, env_config_token, tmpdir, clean_env):
    """
    Creates a config file, and sets the DATAROBOT_CONFIG_FILE environment variable to point
    to it

    Parameters
    ----------
    env_config_endpoint
    env_config_token
    tmpdir
    clean_env

    Returns
    -------
    path : str
        The path to the config file
    """
    fcontent = "endpoint: {}\ntoken: {}".format(env_config_endpoint, env_config_token)
    fpath = tmpdir.join("env_test_config.yaml")
    fpath.write(fcontent)
    clean_env.setenv("DATAROBOT_CONFIG_FILE", str(fpath))
    return str(fpath)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot.models.prime_file import PrimeFile


@pytest.fixture
def downloaded_code():
    return "Copyright DataRobot Code\ndef do_a_thing():\n    return"


@pytest.fixture
def download_response(project_url, prime_file, downloaded_code):
    url = "{}primeFiles/{}/download/".format(project_url, prime_file.id)
    responses.add(responses.GET, url, body=downloaded_code)


def test_future_proof(prime_file_server_data):
    PrimeFile.from_server_data(dict(prime_file_server_data, future="new"))


@responses.activate
def test_get_prime_file(project_url, project_id, prime_file_server_data):
    url = "{}primeFiles/{}/".format(project_url, prime_file_server_data["id"])
    responses.add(
        responses.GET, url, body=json.dumps(prime_file_server_data), content_type="application/json"
    )
    prime_file = PrimeFile.get(project_id, prime_file_server_data["id"])
    assert prime_file.id == prime_file_server_data["id"]
    assert prime_file.project_id == project_id
    assert prime_file.parent_model_id == prime_file_server_data["parentModelId"]
    assert prime_file.model_id == prime_file_server_data["modelId"]
    assert prime_file.ruleset_id == prime_file_server_data["rulesetId"]
    assert prime_file.language == prime_file_server_data["language"]
    assert prime_file.is_valid == prime_file_server_data["isValid"]


@responses.activate
@pytest.mark.usefixtures("download_response")
def test_download(temporary_file, prime_file, downloaded_code):
    prime_file.download(temporary_file)
    with open(temporary_file) as in_f:
        saved_code = in_f.read()
    assert saved_code == downloaded_code



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import mock
import pytest
import responses

from datarobot._experimental.models.enums import UnsupervisedTypeEnum
from datarobot._experimental.models.model import CombinedModel
from datarobot._experimental.models.project import Project
from tests.utils import add_response

_base_url = "https://host_name.com/"
_project_id = "projectId"
_project_a_id = "projectA"
_project_b_id = "projectB"
_combined_model_id = "combinedModelId"

project_a_json = {
    "id": "projectA",
    "projectName": "projectA",
    "unsupervisedMode": True,
}
project_b_json = {
    "id": "projectB",
    "projectName": "projectB",
    "unsupervisedMode": True,
}


@responses.activate
def test_get_segments_models_with_combined_model_id():
    add_response(
        "https://host_name.com/projects/projectA/", project_a_json, method=responses.GET,
    )
    add_response(
        "https://host_name.com/projects/projectB/", project_b_json, method=responses.GET,
    )

    with mock.patch.object(CombinedModel, "get") as mock_combined_model_get:
        with mock.patch.object(Project, "get_models") as mock_project_get_models:
            mock_combined_model_get.return_value = mock.Mock(
                get_segments_info=mock.MagicMock(
                    return_value=[
                        mock.Mock(segment="segment_a", project_id="projectA"),
                        mock.Mock(segment="segment_b", project_id="projectB"),
                    ]
                ),
                id="combinedModelId",
            )
            mock_project_get_models.return_value = ["model"]

            project = Project(id=_project_id)

            # Test with segment keys
            expected_segments_models = [
                {
                    "segment": "segment_a",
                    "project_id": "projectA",
                    "parent_project_id": "projectId",
                    "combined_model_id": "combinedModelId",
                    "models": ["model"],
                },
                {
                    "segment": "segment_b",
                    "project_id": "projectB",
                    "parent_project_id": "projectId",
                    "combined_model_id": "combinedModelId",
                    "models": ["model"],
                },
            ]
            segments_models = project.get_segments_models(_combined_model_id)
            assert segments_models == expected_segments_models


@responses.activate
def test_get_segments_models_without_combined_model_id():
    add_response(
        "https://host_name.com/projects/projectA/", project_a_json, method=responses.GET,
    )
    add_response(
        "https://host_name.com/projects/projectB/", project_b_json, method=responses.GET,
    )

    mock_combined_model = mock.Mock(
        get_segments_info=mock.MagicMock(
            return_value=[
                mock.Mock(segment="segment_a", project_id="projectA"),
                mock.Mock(segment="segment_b", project_id="projectB"),
            ]
        ),
        id="combinedModelId",
    )

    with mock.patch.object(Project, "get_models") as mock_project_get_models:
        mock_project_get_models.return_value = ["model"]

        project = Project(id=_project_id)
        project.get_combined_models = mock.MagicMock(return_value=[mock_combined_model])

        # Test with segment keys
        expected_segments_models = [
            {
                "segment": "segment_a",
                "project_id": "projectA",
                "parent_project_id": "projectId",
                "combined_model_id": "combinedModelId",
                "models": ["model"],
            },
            {
                "segment": "segment_b",
                "project_id": "projectB",
                "parent_project_id": "projectId",
                "combined_model_id": "combinedModelId",
                "models": ["model"],
            },
        ]
        segments_models = project.get_segments_models()
        assert segments_models == expected_segments_models


def test_get_segments_models_without_combined_model_id_fails_with_more_than_one():
    project = Project(id=_project_id)
    project.get_combined_models = mock.MagicMock(return_value=["combined-1", "combined-2"])

    # Test with segment keys
    err = "More than 1 combined_model_id was found, please specify the id that you wish to use."
    with pytest.raises(ValueError, match=err):
        project.get_segments_models()


def test_project_with_no_unsupervised_has_no_clustering():
    mock_client_patch = mock.MagicMock(return_value=mock.Mock(headers={"Location": None}))
    project = Project(id=_project_id)
    project._client.patch = mock_client_patch
    project.from_async = mock.MagicMock()
    project.refresh = mock.MagicMock()

    project.set_target(unsupervised_type="something", autopilot_cluster_list=[1, 2])
    assert mock_client_patch.call_count == 1
    actual_calls = mock_client_patch.call_args
    assert actual_calls[0][0] == "projects/projectId/aim/"
    assert actual_calls[1]["data"] == {
        "target": None,
        "mode": "auto",
        "metric": None,
        "quickrun": True,
    }


def test_project_with_clustering_has_type():
    mock_client_patch = mock.MagicMock(return_value=mock.Mock(headers={"Location": None}))
    project = Project(id=_project_id)
    project._client.patch = mock_client_patch
    project.from_async = mock.MagicMock()
    project.refresh = mock.MagicMock()

    project.set_target(unsupervised_mode=True, unsupervised_type=UnsupervisedTypeEnum.CLUSTERING)
    assert mock_client_patch.call_count == 1
    actual_calls = mock_client_patch.call_args
    assert actual_calls[0][0] == "projects/projectId/aim/"
    assert actual_calls[1]["data"] == {
        "target": None,
        "mode": "auto",
        "metric": None,
        "quickrun": True,
        "unsupervised_mode": True,
        "unsupervised_type": UnsupervisedTypeEnum.CLUSTERING,
    }


def test_project_with_clustering_can_have_autopilot_cluster_list():
    mock_client_patch = mock.MagicMock(return_value=mock.Mock(headers={"Location": None}))
    project = Project(id=_project_id)
    project._client.patch = mock_client_patch
    project.from_async = mock.MagicMock()
    project.refresh = mock.MagicMock()

    project.set_target(
        unsupervised_mode=True,
        unsupervised_type=UnsupervisedTypeEnum.CLUSTERING,
        autopilot_cluster_list=[1, 2, 3],
    )
    assert mock_client_patch.call_count == 1
    actual_calls = mock_client_patch.call_args
    assert actual_calls[0][0] == "projects/projectId/aim/"
    assert actual_calls[1]["data"] == {
        "target": None,
        "mode": "auto",
        "metric": None,
        "quickrun": True,
        "unsupervised_mode": True,
        "unsupervised_type": UnsupervisedTypeEnum.CLUSTERING,
        "autopilot_cluster_list": [1, 2, 3],
    }


@mock.patch("datarobot._experimental.models.project.get_id_from_response")
@mock.patch("datarobot._experimental.models.project.ModelJob")
def test_project_train_datetime_can_send_n_clusters(mock_model_job, mock_get_id_from_reponse):
    mock_client_post = mock.MagicMock(return_value=mock.Mock(headers={"Location": None}))
    project = Project(id=_project_id)
    project._client.post = mock_client_post

    project.train_datetime(
        blueprint_id="blueprint-id", n_clusters=3,
    )
    assert mock_client_post.call_count == 1
    actual_calls = mock_client_post.call_args
    assert actual_calls[0][0] == "projects/projectId/datetimeModels/"
    assert actual_calls[1]["data"] == {
        "blueprint_id": "blueprint-id",
        "n_clusters": 3,
    }



# coding: utf-8
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot.errors import ClientError
from datarobot.helpers import DatetimePartitioning
from datarobot.helpers.partitioning_methods import Backtest
from datarobot.models import ExternalBaselineValidationInfo


@pytest.fixture()
def get_validation_info_url(project):
    return (
        "https://host_name.com/projects/{pid}/externalTimeSeriesBaselineDataValidationJobs/1234/"
    ).format(pid=project.id)


@pytest.fixture
def datetime_partitioning(project):
    return DatetimePartitioning(
        project_id=project.id,
        datetime_partition_column="Date",
        multiseries_id_columns=["Store"],
        forecast_window_start=1,
        forecast_window_end=7,
        holdout_start_date="2014-04-09 00:00:00",
        holdout_end_date="2014-06-15 00:00:00",
        backtests=[
            Backtest(
                validation_start_date="2014-02-01 00:00:00",
                validation_end_date="2014-04-09 00:00:00",
            ),
            Backtest(
                validation_start_date="2013-11-26 00:00:00",
                validation_end_date="2014-02-01 00:00:00",
            ),
        ],
    )


@pytest.fixture
def validate_baseline_payload(datetime_partitioning):
    catalog_version_id = "611c774a36e4bc0388355c03"
    target = "Sales"
    backtests = datetime_partitioning.backtests
    return {
        "catalog_version_id": catalog_version_id,
        "target": target,
        "datetime_partition_column": datetime_partitioning.datetime_partition_column,
        "forecast_window_start": datetime_partitioning.forecast_window_start,
        "forecast_window_end": datetime_partitioning.forecast_window_end,
        "holdout_start_date": datetime_partitioning.holdout_start_date,
        "holdout_end_date": datetime_partitioning.holdout_end_date,
        "backtests": [
            {
                "validation_start_date": backtests[0].validation_start_date,
                "validation_end_date": backtests[0].validation_end_date,
            },
            {
                "validation_start_date": backtests[1].validation_start_date,
                "validation_end_date": backtests[1].validation_end_date,
            },
        ],
        "multiseries_id_columns": datetime_partitioning.multiseries_id_columns,
    }


@pytest.fixture
def validate_external_time_series_baseline_response(
    project, validate_baseline_payload, get_validation_info_url
):
    post_url = (
        "https://host_name.com/projects/{pid}/externalTimeSeriesBaselineDataValidationJobs/"
    ).format(pid=project.id)
    async_url = "https://host_name.com/status/status-id/"

    get_body = validate_baseline_payload.copy()
    get_body.update(
        {
            "project_id": project.id,
            "baseline_validation_job_id": "1234",
            "is_external_baseline_dataset_valid": True,
        }
    )

    responses.add(
        responses.POST,
        post_url,
        body=json.dumps(validate_baseline_payload),
        status=202,
        adding_headers={"Location": async_url},
    )
    responses.add(
        responses.GET,
        async_url,
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": get_validation_info_url},
    )
    responses.add(
        responses.GET,
        get_validation_info_url,
        status=200,
        body=json.dumps(get_body),
        content_type="application/json",
    )


@pytest.fixture()
def validation_info_not_found_response(get_validation_info_url):
    responses.add(
        responses.GET,
        get_validation_info_url,
        status=404,
        content_type="application/json",
        body=json.dumps({"message": "External time series validation job not found."}),
    )


@responses.activate
@pytest.mark.usefixtures("validation_info_not_found_response")
def test_retrieve_validation_info_not_found(project):
    with pytest.raises(ClientError):
        ExternalBaselineValidationInfo.get(project.id, validation_job_id="1234")


@responses.activate
@pytest.mark.usefixtures("validate_external_time_series_baseline_response")
def test_project_validate_external_time_series_baseline(
    project, datetime_partitioning, validate_baseline_payload
):
    validation_info = project.validate_external_time_series_baseline(
        catalog_version_id=validate_baseline_payload["catalog_version_id"],
        target=validate_baseline_payload["target"],
        datetime_partitioning=datetime_partitioning,
    )
    assert validation_info.baseline_validation_job_id == "1234"
    assert validation_info.project_id == project.id
    assert validation_info.is_external_baseline_dataset_valid is True
    assert validation_info.message is None

    keys_to_check = [
        "catalog_version_id",
        "target",
        "datetime_partition_column",
        "multiseries_id_columns",
        "holdout_start_date",
        "holdout_end_date",
        "backtests",
        "forecast_window_start",
        "forecast_window_end",
    ]
    for key in keys_to_check:
        assert getattr(validation_info, key) == validate_baseline_payload[key]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

import datarobot as dr
from datarobot.models.data_engine_query_generator import (
    QueryGeneratorDataset,
    QueryGeneratorSettings,
)

_base_url = "https://host_name.com"
_create_url = "{}/dataEngineQueryGenerators/"
_get_url = "{}/dataEngineQueryGenerators/{}/"

_query_generator_id = "queryGeneratorId"
_query_generator_url = _get_url.format(_base_url, _query_generator_id)


@pytest.fixture
def query_generator_record():
    return {
        "id": "queryGeneratorId",
        "query": "queryText",
        "datasets": [
            {
                "datasetId": "datasetId",
                "datasetVersionId": "datasetVersionId",
                "alias": "datasetAlias",
            }
        ],
        "generatorType": "TimeSeries",
        "generatorSettings": {
            "target": "y",
            "datetimePartitionColumn": "date",
            "multiSeriesIdColumns": ["id"],
            "timeUnit": "DAY",
            "timeStep": 1,
            "defaultNumericAggregationMethod": "sum",
            "defaultCategoricalAggregationMethod": "mostFrequent",
            "defaultTextAggregationMethod": "concatenate",
            "startFromSeriesMinDatetime": True,
            "endToSeriesMaxDatetime": True,
        },
        "generatedQuery": "queryText",
        "inputFeatureTypes": {},
    }


@pytest.fixture
def query_generator_get_json(query_generator_record):
    return json.dumps(query_generator_record)


def test_query_generator_dataset():
    dataset = QueryGeneratorDataset(
        alias="my_dataset", dataset_id="my_dataset_id", dataset_version_id="my_dataset_version_id"
    )
    assert dataset.to_dict() == {
        "alias": "my_dataset",
        "dataset_id": "my_dataset_id",
        "dataset_version_id": "my_dataset_version_id",
    }
    assert dataset.to_payload() == {
        "alias": "my_dataset",
        "datasetId": "my_dataset_id",
        "datasetVersionId": "my_dataset_version_id",
    }


def test_query_generator_settings():
    settings = QueryGeneratorSettings(
        datetime_partition_column="date",
        time_unit="DAY",
        time_step=1,
        default_numeric_aggregation_method="sum",
        default_categorical_aggregation_method="mostFrequent",
        target="y",
        multiseries_id_columns=["id"],
        default_text_aggregation_method="concatenate",
        start_from_series_min_datetime=True,
        end_to_series_max_datetime=True,
    )
    assert settings.to_dict() == {
        "datetime_partition_column": "date",
        "time_unit": "DAY",
        "time_step": 1,
        "default_numeric_aggregation_method": "sum",
        "default_categorical_aggregation_method": "mostFrequent",
        "target": "y",
        "multiseries_id_columns": ["id"],
        "default_text_aggregation_method": "concatenate",
        "start_from_series_min_datetime": True,
        "end_to_series_max_datetime": True,
    }
    assert settings.to_payload() == {
        "datetimePartitionColumn": "date",
        "timeUnit": "DAY",
        "timeStep": 1,
        "defaultNumericAggregationMethod": "sum",
        "defaultCategoricalAggregationMethod": "mostFrequent",
        "target": "y",
        "multiseriesIdColumns": ["id"],
        "defaultTextAggregationMethod": "concatenate",
        "startFromSeriesMinDatetime": True,
        "endToSeriesMaxDatetime": True,
    }


@responses.activate
def test_query_generator_get_valid(query_generator_get_json):
    responses.add(
        responses.GET,
        _query_generator_url,
        body=query_generator_get_json,
        status=200,
        content_type="application/json",
    )
    query_generator = dr.DataEngineQueryGenerator.get(_query_generator_id)

    assert isinstance(query_generator, dr.DataEngineQueryGenerator)
    assert query_generator.id == _query_generator_id
    assert query_generator.datasets == [
        {
            "dataset_id": "datasetId",
            "dataset_version_id": "datasetVersionId",
            "alias": "datasetAlias",
        }
    ]
    assert query_generator.generator_type == "TimeSeries"
    assert query_generator.generator_settings == {
        "target": "y",
        "datetime_partition_column": "date",
        "multi_series_id_columns": ["id"],
        "time_unit": "DAY",
        "time_step": 1,
        "default_numeric_aggregation_method": "sum",
        "default_categorical_aggregation_method": "mostFrequent",
        "default_text_aggregation_method": "concatenate",
        "start_from_series_min_datetime": True,
        "end_to_series_max_datetime": True,
    }
    assert query_generator.query == "queryText"


@pytest.fixture
@responses.activate
def query_generator(query_generator_get_json):
    responses.add(
        responses.POST,
        _create_url.format(_base_url),
        body="",
        status=202,
        adding_headers={"Location": "{}/status/some-status-id/".format(_base_url)},
        content_type="application_json",
    )
    responses.add(
        responses.GET,
        "{}/status/some-status-id/".format(_base_url),
        body="",
        status=303,
        adding_headers={"Location": _query_generator_url},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        _query_generator_url,
        body=query_generator_get_json,
        status=200,
        content_type="application/json",
    )
    query_generator = dr.DataEngineQueryGenerator.create(
        generator_type="TimeSeries",
        datasets=[
            QueryGeneratorDataset(
                alias="datasetAlias", dataset_id="datasetId", dataset_version_id="datasetVersionId"
            )
        ],
        generator_settings=QueryGeneratorSettings(
            datetime_partition_column="date",
            time_unit="DAY",
            time_step=1,
            default_numeric_aggregation_method="sum",
            default_categorical_aggregation_method="mostFrequent",
            target="y",
            multiseries_id_columns=["id"],
            default_text_aggregation_method="concatenate",
            start_from_series_min_datetime=True,
            end_to_series_max_datetime=True,
        ),
    )
    return query_generator


def test_query_generator_create(query_generator):
    assert isinstance(query_generator, dr.DataEngineQueryGenerator)
    assert query_generator.id == _query_generator_id
    assert query_generator.datasets == [
        {
            "dataset_id": "datasetId",
            "dataset_version_id": "datasetVersionId",
            "alias": "datasetAlias",
        }
    ]
    assert query_generator.generator_type == "TimeSeries"
    assert query_generator.generator_settings == {
        "target": "y",
        "datetime_partition_column": "date",
        "multi_series_id_columns": ["id"],
        "time_unit": "DAY",
        "time_step": 1,
        "default_numeric_aggregation_method": "sum",
        "default_categorical_aggregation_method": "mostFrequent",
        "default_text_aggregation_method": "concatenate",
        "start_from_series_min_datetime": True,
        "end_to_series_max_datetime": True,
    }
    assert query_generator.query == "queryText"


@responses.activate
def test_query_generator_create_dataset(query_generator, mock_dataset):
    responses.add(
        responses.POST,
        "{}/dataEngineWorkspaceStates/fromDataEngineQueryGenerator/".format(_base_url),
        body=json.dumps({"workspaceStateId": "someWorkspaceStateId"}),
        status=201,
        content_type="application_json",
    )
    responses.add(
        responses.POST,
        "{}/datasets/fromDataEngineWorkspaceState/".format(_base_url),
        body="",
        status=202,
        adding_headers={"Location": "{}/status/some-status-id/".format(_base_url)},
        content_type="application_json",
    )
    dataset_url = "{}/datasets/{}".format(_base_url, mock_dataset["datasetId"])
    responses.add(
        responses.GET,
        "{}/status/some-status-id/".format(_base_url),
        body="",
        status=303,
        adding_headers={"Location": dataset_url},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        dataset_url,
        body=json.dumps(mock_dataset),
        status=200,
        content_type="application/json",
    )
    dataset = query_generator.create_dataset()
    assert isinstance(dataset, dr.Dataset)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

from mock import mock
import pytest
import responses

from datarobot import DatetimeModel
from datarobot.enums import AnomalyAssessmentStatus
from datarobot.models.anomaly_assessment import (
    AnomalyAssessmentExplanations,
    AnomalyAssessmentPredictionsPreview,
    AnomalyAssessmentRecord,
)


@pytest.fixture
def project_id():
    return "5ece5927962d741e19c2febb"


@pytest.fixture
def model_id():
    return "607d6544823fa548132dc257"


@pytest.fixture
def completed_record_id():
    return "607d657551a4bef68dd29d2f"


@pytest.fixture
def anomaly_assessment_single_record_response(unittest_endpoint):
    latest_loc = (
        "{}/projects/5ece5927962d741e19c2febb/anomalyAssessmentRecords/"
        "607d657551a4bef68dd29d2f/explanations/"
        "?endDate=2015-11-02T00%3A00%3A00.000000Z&pointsCount=2".format(unittest_endpoint)
    )
    return {
        "count": 1,
        "next": None,
        "data": [
            {
                "status": "completed",
                "seriesId": "S1F03YZM",
                "endDate": "2015-11-02T00:00:00.000000Z",
                "startDate": "2015-10-07T00:00:00.000000Z",
                "recordId": "607d657551a4bef68dd29d2f",
                "projectId": "5ece5927962d741e19c2febb",
                "backtest": "holdout",
                "source": "training",
                "latestExplanationsLocation": latest_loc,
                "deleteLocation": "{}/projects/5ece5927962d741e19c2febb/"
                "anomalyAssessmentRecords/607d657551a4bef68dd29d2f/".format(unittest_endpoint),
                "predictionThreshold": 0.001152261490359791,
                "statusDetails": "insight data is available.",
                "previewLocation": "{}/projects/5ece5927962d741e19c2febb/"
                "anomalyAssessmentRecords/607d657551a4bef68dd29d2f/predictionsPreview/".format(
                    unittest_endpoint
                ),
                "modelId": "607d6544823fa548132dc257",
            },
        ],
        "previous": None,
    }


@pytest.fixture
def anomaly_assessment_records_response(unittest_endpoint):
    return {
        "count": 2,
        "next": "{}/projects/5ece5927962d741e19c2febb/anomalyAssessmentRecords/"
        "?limit=2&offset=2".format(unittest_endpoint),
        "data": [
            {
                "status": "noData",
                "seriesId": "S1F03YZM",
                "endDate": None,
                "startDate": None,
                "recordId": "607edabbe8c1f4b282d29c20",
                "projectId": "5ece5927962d741e19c2febb",
                "backtest": 0,
                "source": "validation",
                "latestExplanationsLocation": None,
                "deleteLocation": "{}/projects/5ece5927962d741e19c2febb/anomalyAssessmentRecords/"
                "607edabbe8c1f4b282d29c20/".format(unittest_endpoint),
                "predictionThreshold": None,
                "statusDetails": "There is no data for specified parameters.",
                "previewLocation": None,
                "modelId": "607d6544823fa548132dc257",
            },
            {
                "status": "completed",
                "seriesId": "S1F0EGMT",
                "endDate": "2015-11-02T00:00:00.000000Z",
                "startDate": "2015-10-07T00:00:00.000000Z",
                "recordId": "607d657551a4bef68dd29d2f",
                "projectId": "5ece5927962d741e19c2febb",
                "backtest": 0,
                "source": "validation",
                "latestExplanationsLocation": "{}/projects/5ece5927962d741e19c2febb/"
                "anomalyAssessmentRecords/607d657551a4bef68dd29d2f/"
                "explanations/?endDate=2015-11-02T00%3A00%3A00.000000Z&pointsCount=2".format(
                    unittest_endpoint
                ),
                "deleteLocation": "{}/projects/5ece5927962d741e19c2febb/anomalyAssessmentRecords/"
                "607d657551a4bef68dd29d2f/".format(unittest_endpoint),
                "predictionThreshold": 0.001152261490359791,
                "statusDetails": "insight data is available.",
                "previewLocation": "{}/projects/5ece5927962d741e19c2febb/anomalyAssessmentRecords/"
                "607d657551a4bef68dd29d2f/predictionsPreview/".format(unittest_endpoint),
                "modelId": "607d6544823fa548132dc257",
            },
        ],
        "previous": None,
    }


@pytest.fixture
def anomaly_explanations_response():
    return {
        "count": 2,
        "shapBaseValue": 0.01,
        "endDate": "2015-11-02T00:00:00.000000Z",
        "seriesId": "S1F0EGMT",
        "startDate": "2015-11-02T00:00:00.000000Z",
        "recordId": "607d657551a4bef68dd29d2f",
        "projectId": "5ece5927962d741e19c2febb",
        "source": "validation",
        "data": [
            {
                "shapExplanation": [
                    {
                        "featureValue": "75929023.55426522",
                        "strength": 2.730025962114207e-07,
                        "feature": "attribute1 (35 day std)",
                    },
                    {
                        "featureValue": "343016.23529411765",
                        "strength": 2.5279249798097965e-07,
                        "feature": "attribute6 (35 day mean)",
                    },
                ],
                "timestamp": "2015-10-31T00:00:00.000000Z",
                "prediction": 0.001152261490359791,
            },
            {"shapExplanation": [], "timestamp": "2015-11-02T00:00:00.000000Z", "prediction": 0.4},
        ],
        "backtest": 0,
        "modelId": "607d6544823fa548132dc257",
    }


@pytest.fixture()
def anomaly_predictions_preview_response():
    return {
        "seriesId": "S1F0EGMT",
        "startDate": "2015-10-07T00:00:00.000000Z",
        "endDate": "2015-11-02T00:00:00.000000Z",
        "recordId": "607d657551a4bef68dd29d2f",
        "projectId": "5ece5927962d741e19c2febb",
        "source": "validation",
        "previewBins": [
            {
                "avgPredicted": 0.4,
                "startDate": "2015-10-07T00:00:00.000000Z",
                "frequency": 2,
                "endDate": "2015-11-03T00:00:00.000000Z",
                "maxPredicted": 0.5,
            }
        ],
        "backtest": 0,
        "modelId": "607d6544823fa548132dc257",
    }


def check_metadata(record, project_id, model_id):
    assert record.project_id == project_id
    assert record.model_id == model_id
    assert record.series_id
    assert record.record_id
    assert record.source == "validation"
    assert record.backtest == 0


@responses.activate
def test_get_anomaly_assessment_records(
    anomaly_assessment_records_response,
    anomaly_predictions_preview_response,
    anomaly_explanations_response,
    unittest_endpoint,
    project_id,
    model_id,
    completed_record_id,
):

    responses.add(
        responses.GET,
        "{}/projects/{}/anomalyAssessmentRecords/?limit=100&offset=0&modelId={}".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_assessment_records_response),
    )
    responses.add(
        responses.GET,
        "{}/projects/{}/anomalyAssessmentRecords/{}/explanations/?"
        "endDate=2015-11-02T00%3A00%3A00.000000Z&pointsCount=2".format(
            unittest_endpoint, project_id, completed_record_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_explanations_response),
    )
    responses.add(
        responses.GET,
        "{}/projects/{}/anomalyAssessmentRecords/{}/explanations/?"
        "startDate=2015-10-07T00%3A00%3A00.000000Z&"
        "endDate=2015-11-03T00%3A00%3A00.000000Z".format(
            unittest_endpoint, project_id, completed_record_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_explanations_response),
    )
    responses.add(
        responses.GET,
        "{}/projects/{}/anomalyAssessmentRecords/{}/predictionsPreview/".format(
            unittest_endpoint, project_id, completed_record_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_predictions_preview_response),
    )
    responses.add(
        responses.DELETE,
        "{}/projects/{}/anomalyAssessmentRecords/{}/".format(
            unittest_endpoint, project_id, completed_record_id
        ),
        status=204,
        content_type="application/json",
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    records = model.get_anomaly_assessment_records()
    no_data_record = records[0]
    completed_record = records[1]
    assert no_data_record.series_id == "S1F03YZM"
    assert no_data_record.status == AnomalyAssessmentStatus.NO_DATA
    assert no_data_record.latest_explanations_location is None
    assert no_data_record.preview_location is None
    assert no_data_record.start_date is None
    assert no_data_record.end_date is None
    assert no_data_record.prediction_threshold is None
    assert completed_record.series_id == "S1F0EGMT"
    assert completed_record.status == AnomalyAssessmentStatus.COMPLETED
    for record in records:
        check_metadata(record, project_id, model_id)
        assert record.delete_location
    explanations = completed_record.get_latest_explanations()
    assert isinstance(explanations, AnomalyAssessmentExplanations)
    preview = completed_record.get_predictions_preview()
    assert isinstance(preview, AnomalyAssessmentPredictionsPreview)
    assert preview.series_id == "S1F0EGMT"
    check_metadata(preview, project_id, model_id)
    assert preview.preview_bins == [
        {
            "avg_predicted": 0.4,
            "max_predicted": 0.5,
            "start_date": "2015-10-07T00:00:00.000000Z",
            "end_date": "2015-11-03T00:00:00.000000Z",
            "frequency": 2,
        }
    ]
    check_metadata(explanations, project_id, model_id)
    assert explanations.shap_base_value == 0.01
    assert explanations.count == 2
    assert explanations.series_id == "S1F0EGMT"
    assert explanations.data == [
        {
            "shap_explanation": [
                {
                    "feature_value": "75929023.55426522",
                    "strength": mock.ANY,
                    "feature": "attribute1 (35 day std)",
                },
                {
                    "feature_value": "343016.23529411765",
                    "strength": mock.ANY,
                    "feature": "attribute6 (35 day mean)",
                },
            ],
            "timestamp": "2015-10-31T00:00:00.000000Z",
            "prediction": mock.ANY,
        },
        {
            "shap_explanation": [],
            "timestamp": "2015-11-02T00:00:00.000000Z",
            "prediction": mock.ANY,
        },
    ]
    regions = preview.find_anomalous_regions(max_prediction_threshold=0.5)
    assert len(regions) == 1
    explanations = completed_record.get_explanations_data_in_regions(
        regions, prediction_threshold=0.4
    )
    assert len(explanations["explanations"]) == 1
    assert explanations["shap_base_value"] == 0.01
    assert completed_record.delete() is None


@responses.activate
def test_initialize_anomaly_assessment(
    unittest_endpoint, project_id, model_id, anomaly_assessment_single_record_response
):

    backtest = "holdout"
    source = "training"
    series_id = "S1F03YZM"
    responses.add(
        responses.POST,
        "{}/projects/{}/models/{}/anomalyAssessmentInitialization/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=202,
        content_type="application/json",
        body=json.dumps({"seriesId": series_id, "backtest": backtest, "source": source}),
        adding_headers={"Location": "{}/status/status-id/".format(unittest_endpoint)},
    )
    responses.add(
        responses.GET,
        "{}/status/status-id/".format(unittest_endpoint),
        status=303,
        body="",
        content_type="application/json",
        adding_headers={
            "Location": "{}/projects/{}/anomalyAssessmentRecords/?seriesId={}"
            "&backtest={}&source={}".format(
                unittest_endpoint, project_id, model_id, series_id, backtest, source
            )
        },
    )
    record_resp_url = (
        "{}/projects/{}/anomalyAssessmentRecords/"
        "?seriesId={}&backtest={}&source={}".format(
            unittest_endpoint, project_id, model_id, series_id, backtest, source
        )
    )
    responses.add(
        responses.GET,
        "{}/status/status-id/".format(unittest_endpoint),
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": record_resp_url},
    )
    responses.add(
        responses.GET,
        record_resp_url,
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_assessment_single_record_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    new_record = model.initialize_anomaly_assessment(backtest, source, series_id=series_id)
    assert isinstance(new_record, AnomalyAssessmentRecord)
    assert new_record.backtest == backtest
    assert new_record.series_id == series_id
    assert new_record.source == source
    assert new_record.model_id == model_id
    assert new_record.project_id == project_id



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot import errors, FeatureAssociationFeaturelists


@responses.activate
def test_get(
    project_id, feature_association_featurelists_url, feature_association_featurelists_data
):
    responses.add(
        responses.GET,
        feature_association_featurelists_url,
        status=200,
        content_type="application/json",
        body=json.dumps(feature_association_featurelists_data),
    )
    fam_featurelists = FeatureAssociationFeaturelists.get(project_id)

    assert len(fam_featurelists.featurelists) == 2
    for fam_featurelist in fam_featurelists.featurelists:
        assert set(fam_featurelist.keys()) == {"title", "has_fam", "featurelist_id"}


@responses.activate
def test_get__not_found(project_id, feature_association_featurelists_url):
    responses.add(responses.GET, feature_association_featurelists_url, status=404)
    with pytest.raises(errors.ClientError, match="404 client error: "):
        FeatureAssociationFeaturelists.get(project_id)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import datetime
import unittest

import pandas as pd
import pytest
import responses

from datarobot import errors
from datarobot.client import Client, set_client
from datarobot.errors import DataRobotDeprecationWarning
from datarobot.utils import (
    dataframe_to_buffer,
    deprecation,
    from_api,
    get_id_from_response,
    is_urlsource,
    parse_time,
    recognize_sourcedata,
    to_api,
)
from datarobot.utils.sourcedata import list_of_records_to_buffer
from tests.test_helpers import fixture_file_path
from tests.utils import warns


class TestDataframeSerialization(unittest.TestCase):
    def test_list_of_records_to_buffer(self):
        records = [{"a": "b", "c": "d"}, {"a": "x", "c": "y"}]
        expected = "a,c\r\nb,d\r\nx,y\r\n"
        buff = list_of_records_to_buffer(records)
        self.assertEqual(buff.read(), expected)

    def test_no_index_please(self):
        df = pd.DataFrame({"a": range(100), "b": range(100)})
        buff = dataframe_to_buffer(df)
        readback = pd.read_csv(buff)
        self.assertEqual(readback.columns.tolist(), ["a", "b"])

    def test_parse_time(self):
        self.assertEqual("BAD TIME", parse_time("BAD TIME"))  # returns value
        test_string_time = datetime.datetime.now().isoformat()
        self.assertIsInstance(test_string_time, str)
        self.assertIsInstance(parse_time(test_string_time), datetime.datetime)

    @responses.activate
    @pytest.mark.usefixtures("no_client_version_check")
    def test_get_id_from_response_location_header(self):
        responses.add(
            responses.POST,
            "http://nothing/",
            body="",
            adding_headers={"Location": "http://nothing/f-id/"},
        )
        client = set_client(Client(token="no_matter", endpoint="http://nothing"))
        resp = client.post("")
        self.assertEqual(get_id_from_response(resp), "f-id")

    def test_csv_buffer_is_quoted(self):
        """"
        This tests that dataframes are correctly quoted when saved back to csv, so
        that columns containing line endings won't be interpreted as multiple rows
        See: PRED-3203
        """
        df = pd.DataFrame({"A": ["test\rtest", 2], "B": ["test\n\n\n\r\ntest", 0.75]})
        buff = dataframe_to_buffer(df)
        readback = pd.read_csv(buff)
        assert df.shape[0] == readback.shape[0]


class TestFromAPI(unittest.TestCase):
    def test_nested_list_of_objects_all_changed(self):
        source = {"oneFish": [{"twoFish": "redFish"}, {"blueFish": "noFish"}]}
        result = from_api(source)
        inner = result["one_fish"]
        self.assertEqual(inner[0]["two_fish"], "redFish")
        self.assertEqual(inner[1]["blue_fish"], "noFish")

    def test_nested_objects_all_changed(self):
        source = {"oneFish": {"twoFish": "redFish"}}

        result = from_api(source)
        self.assertEqual(result["one_fish"]["two_fish"], "redFish")


def test_from_api_filter_nones():
    data_in = {
        "someValue": {"nestedNone": None, "nestedNotNone": ""},
        "listNone": [{"someValue": None, "someNotNone": ""}],
        "otherNone": None,
    }
    data_out = {"some_value": {"nested_not_none": ""}, "list_none": [{"some_not_none": ""}]}
    assert data_out == from_api(data_in)


def test_from_api_no_filter_nones():
    data_in = {
        "someValue": {"nestedNone": None, "nestedNotNone": ""},
        "listNone": [{"someValue": None, "someNotNone": ""}],
        "otherNone": None,
    }
    data_out = {
        "some_value": {"nested_none": None, "nested_not_none": ""},
        "list_none": [{"some_value": None, "some_not_none": ""}],
        "other_none": None,
    }
    assert data_out == from_api(data_in, keep_null_keys=True)


def test_from_api_keep_nones():
    data_in = {
        "someValue": {"nestedNone": None, "nestedNotNone": "", "levelTwo": {"levelThree": None}},
        "otherNone": None,
        "keepLevelOne": None,
    }
    data_out = {
        "some_value": {
            "nested_not_none": "",
            "nested_none": None,
            "level_two": {"level_three": None},
        },
        "keep_level_one": None,
    }
    assert data_out == from_api(
        data_in,
        keep_attrs=["some_value.nested_none", "some_value.level_two.level_three", "keep_level_one"],
    )


def test_from_api_multiple_underscore_ignored():
    in_data = {"samplePct__gte": 64}
    expected = {"sample_pct__gte": 64}
    assert from_api(in_data) == expected


def test_to_api_multiple_underscore_ignored():
    in_data = {"sample_pct__gte": 64}
    expected = {"samplePct__gte": 64}
    assert to_api(in_data) == expected


def test_from_api_single_letter():
    in_data = {"iAmAKey": "test"}
    expected = {"i_am_a_key": "test"}
    assert from_api(in_data) == expected


def test_to_api_single_letter():
    in_data = {"i_am_a_key": "test"}
    expected = {"iAmAKey": "test"}
    assert to_api(in_data) == expected


@deprecation.deprecated(deprecated_since_version="v0.1.2", will_remove_version="v1.2.3")
def bar(arg1, arg2):
    """
    A dummy function to use for testing the deprecation mechanisms

    Parameters
    ----------
    arg1 : int
    arg2 : int

    Returns
    -------
    total : int
        arg1 + arg2

    """
    return arg1 + arg2


@deprecation.deprecated(
    deprecated_since_version="v0.1.2",
    will_remove_version="v1.2.3",
    message="Use of `bar` is recommended instead.",
)
def foo(arg1, arg2):
    """
    A dummy function to use for testing the deprecation mechanisms

    Parameters
    ----------
    arg1 : int
    arg2 : int

    Returns
    -------
    total : int
        foo + bar

    """
    return arg1 + arg2


def test_deprecation_with_message(known_warning):
    foo(1, 2)
    w_msg = known_warning[0].message.args[0]
    assert (
        w_msg == "`foo` has been deprecated in `v0.1.2`, will be removed "
        "in `v1.2.3`. Use of `bar` is recommended instead."
    )


def test_deprecation_message(known_warning):
    bar(1, 2)
    w_instance = known_warning[0].message
    assert isinstance(w_instance, DataRobotDeprecationWarning)


class TestSourcedataUtils(unittest.TestCase):
    def setUp(self):
        self.default_fname = "predict.csv"

    def test_recognize_sourcedata_passed_dataframe(self):
        df = pd.DataFrame({"a": range(100), "b": range(100)})
        kwargs = recognize_sourcedata(df, self.default_fname)
        self.assertTrue("filelike" in kwargs)
        self.assertEqual(kwargs.get("fname"), self.default_fname)
        buff = kwargs["filelike"]
        readback = pd.read_csv(buff)
        self.assertEqual(readback.columns.tolist(), ["a", "b"])

    def test_recognize_sourcedata_passed_filelike(self):
        path = fixture_file_path("synthetic-100.csv")
        with open(path, "rb") as fd:
            kwargs = recognize_sourcedata(fd, self.default_fname)
            self.assertTrue(kwargs.get("filelike") is fd)
            self.assertEqual(kwargs.get("fname"), self.default_fname)

    def test_recognize_sourcedata_passed_filepath(self):
        file_path = fixture_file_path("synthetic-100.csv")
        kwargs = recognize_sourcedata(file_path, self.default_fname)
        self.assertEqual(kwargs.get("file_path"), file_path)
        self.assertEqual(kwargs.get("fname"), "synthetic-100.csv")

    def test_recognize_sourcedata_passed_content(self):
        content = b"a,b,c\n" + b"1,2,3\n" * 100
        kwargs = recognize_sourcedata(content, self.default_fname)
        self.assertEqual(kwargs.get("content"), content)
        self.assertEqual(kwargs.get("fname"), self.default_fname)

    def test_is_urlsource_passed_true(self):
        result = is_urlsource("http://path_to_urlsource")
        self.assertTrue(result)

    def test_is_urlsource_file_passed_true(self):
        result = is_urlsource("file://path_to_urlsource")
        self.assertTrue(result)

    def test_is_urlsource_file_passed_false(self):
        result = is_urlsource("filename")
        self.assertFalse(result)

    def test_is_urlsource_passed_false(self):
        result = is_urlsource("not_a_path_to_urlsource")
        self.assertFalse(result)

    def test_is_urlsource_s3_passed_true(self):
        result = is_urlsource("s3://path/to/a/bucket/data.csv")
        self.assertTrue(result)

    def test_fatfingered_filepath_raises(self):
        content = b"/home/datarobot/mistypefilepath.csv"
        with self.assertRaises(errors.InputNotUnderstoodError):
            recognize_sourcedata(content, self.default_fname)


def test_resource():
    """Python 3 ResourceWarning of implicitly closed files."""
    with warns():
        open(__file__)


class TestImports(unittest.TestCase):
    def test_rename(self):
        """datarobot_sdk rename."""
        with self.assertRaises(ImportError):
            import datarobot_sdk  # noqa



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import csv
from datetime import datetime
import json
import os

from pandas import DataFrame
from pandas.testing import assert_frame_equal
import pytest
import responses

from datarobot import DatetimeModel, enums, Model
from datarobot.errors import ClientError
from tests.utils import request_body_to_json


def test_cannot_train_by_samplepct(datetime_model):
    with pytest.raises(NotImplementedError) as exc_info:
        datetime_model.train()
    assert "use train_datetime instead" in str(exc_info.value)


def test_cannot_frozen_by_samplepct(datetime_model):
    with pytest.raises(NotImplementedError) as exc_info:
        datetime_model.request_frozen_model()
    assert "use request_frozen_datetime_model instead" in str(exc_info.value)


def _validate_datetime_model(mod):
    assert mod.sample_pct is None

    training_info = mod.training_info
    expected_keys = {
        "holdout_training_start_date",
        "holdout_training_duration",
        "holdout_training_row_count",
        "holdout_training_end_date",
        "prediction_training_start_date",
        "prediction_training_duration",
        "prediction_training_row_count",
        "prediction_training_end_date",
    }
    assert set(training_info.keys()) == expected_keys
    for key in training_info:
        if "date" in key:
            value = training_info[key]
            assert isinstance(value, datetime)
            assert value.tzname() in ["UTC", "Z"]
    assert mod.data_selection_method in ["rowCount", "duration", "selectedDateRange"]
    assert mod.time_window_sample_pct is None or 0 < mod.time_window_sample_pct < 100
    assert mod.holdout_score is None
    assert mod.holdout_status == "COMPLETED"
    train_dates = [mod.training_start_date, mod.training_end_date]
    for date in train_dates:
        if date is not None:
            assert isinstance(date, datetime)
            assert date.tzname() in ["UTC", "Z"]
    backtests = mod.backtests
    expected_keys = {
        "index",
        "score",
        "status",
        "training_start_date",
        "training_duration",
        "training_row_count",
        "training_end_date",
    }
    for bt in backtests:
        assert set(bt.keys()) == expected_keys
        for key in expected_keys:
            if "date" in key and bt[key] is not None:
                date = bt[key]
                assert isinstance(date, datetime)
                assert date.tzname() in ["UTC", "Z"]
    fdw_keys = {
        "effective_feature_derivation_window_start",
        "effective_feature_derivation_window_end",
        "forecast_window_start",
        "forecast_window_end",
        "windows_basis_unit",
    }
    for key in fdw_keys:
        assert hasattr(mod, key)


@responses.activate
def test_retrieve_model(datetime_model_server_data, project_url):
    project_id = datetime_model_server_data["projectId"]
    model_id = datetime_model_server_data["id"]

    responses.add(
        responses.GET,
        "{}datetimeModels/{}/".format(project_url, datetime_model_server_data["id"]),
        json=datetime_model_server_data,
    )

    mod = DatetimeModel.get(project_id, model_id)
    _validate_datetime_model(mod)


@responses.activate
def test_retrieve_model_with_startend_date(datetime_start_end_model_server_data, project_url):
    project_id = datetime_start_end_model_server_data["projectId"]
    model_id = datetime_start_end_model_server_data["id"]

    responses.add(
        responses.GET,
        "{}datetimeModels/{}/".format(project_url, datetime_start_end_model_server_data["id"]),
        json=datetime_start_end_model_server_data,
    )

    mod = DatetimeModel.get(project_id, model_id)
    _validate_datetime_model(mod)


@responses.activate
def test_list_models(
    datetime_model_server_data, datetime_start_end_model_server_data, datetime_project, project_url
):
    json = {
        "previous": None,
        "next": None,
        "count": 2,
        "data": [datetime_model_server_data, datetime_start_end_model_server_data],
    }
    responses.add(responses.GET, "{}datetimeModels/".format(project_url), json=json)
    mods = datetime_project.get_datetime_models()
    assert len(mods) == 2
    for mod in mods:
        _validate_datetime_model(mod)


@responses.activate
def test_train_from_project(datetime_project, project_url, datetime_model_job_server_data):
    model_job_loc = "{}modelJobs/1/".format(project_url)
    responses.add(
        responses.POST,
        "{}datetimeModels/".format(project_url),
        body="",
        adding_headers={"Location": model_job_loc},
    )
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_project.train_datetime("bp-id")

    assert mod_job.id == int(datetime_model_job_server_data["id"])
    assert mod_job.status == datetime_model_job_server_data["status"]
    assert mod_job.processes == datetime_model_job_server_data["processes"]
    assert mod_job.sample_pct is None


@responses.activate
def test_train_from_model(datetime_model, project_url, datetime_model_job_server_data):
    model_job_loc = "{}modelJobs/1/".format(project_url)
    responses.add(
        responses.POST,
        "{}datetimeModels/".format(project_url),
        body="",
        adding_headers={"Location": model_job_loc},
    )
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_model.train_datetime("bp-id")

    assert mod_job.id == int(datetime_model_job_server_data["id"])
    assert mod_job.status == datetime_model_job_server_data["status"]
    assert mod_job.processes == datetime_model_job_server_data["processes"]
    assert mod_job.sample_pct is None


@responses.activate
def test_train_datetime_sampling_method(
    datetime_model, project_url, datetime_model_job_server_data
):
    model_job_loc = "{}modelJobs/1/".format(project_url)
    responses.add(
        responses.POST,
        "{}datetimeModels/".format(project_url),
        body="",
        adding_headers={"Location": model_job_loc},
    )
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_model.train_datetime(
        training_duration="P1Y", time_window_sample_pct=40, sampling_method="latest"
    )
    assert mod_job.id == int(datetime_model_job_server_data["id"])

    payload = request_body_to_json(responses.calls[0].request)
    assert payload.get("timeWindowSamplePct") == 40
    assert payload.get("samplingMethod") == "latest"
    assert payload.get("trainingDuration") == "P1Y"


@responses.activate
def train_request_frozen_datetime(datetime_model, project_url, datetime_model_job_server_data):
    model_job_loc = "{}modelJobs/{}/".format(project_url, datetime_model_job_server_data["id"])
    responses.add(
        responses.POST,
        "{}frozenDatetimeModels/".format(project_url),
        body="",
        status=202,
        adding_headers={"Location": model_job_loc},
    )
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_model.request_frozen_datetime_model(training_row_count=800)

    assert mod_job.id == int(datetime_model_job_server_data["id"])
    assert mod_job.status == datetime_model_job_server_data["status"]
    assert mod_job.processes == datetime_model_job_server_data["processes"]
    assert mod_job.sample_pct is None


@responses.activate
def train_request_frozen_datetime_with_dates(
    datetime_model, project_url, datetime_model_job_server_data
):
    model_job_loc = "{}modelJobs/{}/".format(project_url, datetime_model_job_server_data["id"])
    responses.add(
        responses.POST,
        "{}frozenDatetimeModels/".format(project_url),
        body="",
        status=202,
        adding_headers={"Location": model_job_loc},
    )
    responses.add(responses.GET, model_job_loc, json=datetime_model_job_server_data)

    mod_job = datetime_model.request_frozen_datetime_model(
        training_start_date=datetime.now(),
        training_end_date=datetime.now(),
        time_window_sample_pct=40,
        sampling_method="latest",
    )

    payload = request_body_to_json(responses.calls[0].request)
    assert payload.get("timeWindowSamplePct") == 40
    assert payload.get("samplingMethod") == "latest"

    assert mod_job.id == int(datetime_model_job_server_data["id"])
    assert mod_job.status == datetime_model_job_server_data["status"]
    assert mod_job.processes == datetime_model_job_server_data["processes"]
    assert mod_job.sample_pct is None


@responses.activate
def test_request_frozen_datetime_with_end_date_and_duration(
    datetime_model, project_url, datetime_model_job_server_data
):
    # Here we test a bug which was uncovered by accident - when training_end_date and no
    # training_start_date is passed in, we should pass the request to API, and it should fail on
    # BE side.
    responses.add(
        responses.POST, "{}frozenDatetimeModels/".format(project_url), body="", status=422,
    )

    with pytest.raises(ClientError):
        datetime_model.request_frozen_datetime_model(
            training_duration="some duration",
            training_end_date=datetime.now(),
            time_window_sample_pct=40,
            sampling_method="latest",
        )


@responses.activate
def test_score_backtests(base_job_running_server_data, datetime_model, project_url):
    job_body = dict(base_job_running_server_data, jobType=enums.JOB_TYPE.MODEL)
    job_loc = "{}jobs/{}/".format(project_url, job_body["id"])
    responses.add(
        responses.POST,
        "{}datetimeModels/{}/backtests/".format(project_url, datetime_model.id),
        body="",
        adding_headers={"Location": job_loc},
        status=202,
    )
    responses.add(responses.GET, job_loc, json=job_body)

    job = datetime_model.score_backtests()
    assert job.id == int(job_body["id"])
    assert job.status == job_body["status"]


def test_datetimemodel_dont_cross_validate(datetime_model):
    with pytest.raises(NotImplementedError):
        datetime_model.cross_validate()


@responses.activate
def test_compute_series_accuracy(base_job_running_server_data, datetime_model, project_url):
    job_body = dict(base_job_running_server_data, jobType=enums.JOB_TYPE.MODEL)
    job_loc = "{}jobs/{}/".format(project_url, job_body["id"])
    responses.add(
        responses.POST,
        "{}datetimeModels/{}/multiseriesScores/".format(project_url, datetime_model.id),
        body="",
        adding_headers={"Location": job_loc},
        status=202,
    )
    responses.add(responses.GET, job_loc, json=job_body)

    job = datetime_model.compute_series_accuracy()
    assert job.id == int(job_body["id"])
    assert job.status == job_body["status"]


@responses.activate
def test_compute_series_accuracy_all_series(
    base_job_running_server_data, datetime_model, project_url
):
    job_body = dict(base_job_running_server_data, jobType=enums.JOB_TYPE.MODEL)
    job_loc = "{}jobs/{}/".format(project_url, job_body["id"])
    responses.add(
        responses.POST,
        "{}datetimeModels/{}/multiseriesScores/".format(project_url, datetime_model.id),
        body=json.dumps({"computeAllSeries": True}),
        adding_headers={"Location": job_loc},
        status=202,
    )
    responses.add(responses.GET, job_loc, json=job_body)

    job = datetime_model.compute_series_accuracy(compute_all_series=True)
    assert job.id == int(job_body["id"])
    assert job.status == job_body["status"]


@responses.activate
def test_download_series_accuracy_as_csv(datetime_model, project_url, tmpdir):
    response_data = {
        u"querySeriesCount": 2,
        u"previous": None,
        u"data": [
            {
                u"multiseriesId": u"5c58975a211c0a1698093f25",
                u"validationScore": None,
                u"backtestingScore": None,
                u"rowCount": 365,
                u"multiseriesValues": [u"1"],
                u"holdoutScore": None,
                u"duration": u"P1Y0M0D",
            },
            {
                u"multiseriesId": u"5c58975a211c0a1698093f26",
                u"validationScore": None,
                u"backtestingScore": None,
                u"rowCount": 329,
                u"multiseriesValues": [u"2"],
                u"holdoutScore": None,
                u"duration": u"P0Y10M26D",
            },
        ],
        u"totalSeriesCount": 2,
        u"next": None,
    }
    url = "{}datetimeModels/{}/multiseriesScores/".format(project_url, datetime_model.id)
    responses.add(responses.GET, url, body=json.dumps(response_data), status=200)
    local_csv_filepath = tmpdir.join("series_accuracy.csv")
    datetime_model.download_series_accuracy_as_csv(
        local_csv_filepath.strpath,
        offset=1,
        limit=2,
        metric="asdf",
        multiseries_value="asdf",
        order_by="asdf",
        reverse=True,
    )

    # Check correct params were used
    query_params = {"offset=1", "limit=2", "metric=asdf", "multiseriesValue=asdf", "orderBy=-asdf"}
    for query_param in query_params:
        assert query_param in responses.calls[0].request.url

    assert os.path.isfile(local_csv_filepath.strpath)
    with open(local_csv_filepath.strpath) as f:
        csv_output = csv.reader(f)
        # assert headers
        correct_header_list = [
            "multiseriesId",
            "validationScore",
            "backtestingScore",
            "rowCount",
            "multiseriesValues",
            "holdoutScore",
            "duration",
        ]
        assert sorted(next(csv_output)) == sorted(correct_header_list)
        for line in csv_output:
            assert len(line) == 7


@responses.activate
def test_download_series_accuracy_as_dataframe(datetime_model, project_url):
    response_data = {
        u"querySeriesCount": 2,
        u"previous": None,
        u"data": [
            {
                u"multiseriesId": u"5c58975a211c0a1698093f25",
                u"validationScore": None,
                u"backtestingScore": None,
                u"rowCount": 365,
                u"multiseriesValues": [u"1"],
                u"holdoutScore": None,
                u"duration": u"P1Y0M0D",
            },
            {
                u"multiseriesId": u"5c58975a211c0a1698093f26",
                u"validationScore": None,
                u"backtestingScore": None,
                u"rowCount": 329,
                u"multiseriesValues": [u"2"],
                u"holdoutScore": None,
                u"duration": u"P0Y10M26D",
            },
        ],
        u"totalSeriesCount": 2,
        u"next": None,
    }
    url = "{}datetimeModels/{}/multiseriesScores/".format(project_url, datetime_model.id)
    responses.add(responses.GET, url, body=json.dumps(response_data), status=200)
    df = datetime_model.get_series_accuracy_as_dataframe(
        offset=1, limit=2, metric="asdf", multiseries_value="asdf", order_by="asdf", reverse=True
    )

    # Check correct params were used
    query_params = {"offset=1", "limit=2", "metric=asdf", "multiseriesValue=asdf", "orderBy=-asdf"}
    for query_param in query_params:
        assert query_param in responses.calls[0].request.url

    assert_frame_equal(df, DataFrame(response_data["data"]))


MONO_INC_FL_ID = "5a9975db962d74221ba38dc9"
MONO_DEC_FL_ID = "5a9975db962d74221ba38dca"


@pytest.mark.parametrize(
    "featurelists",
    [
        (MONO_INC_FL_ID, MONO_DEC_FL_ID),
        (enums.MONOTONICITY_FEATURELIST_DEFAULT, enums.MONOTONICITY_FEATURELIST_DEFAULT),
        (None, None),
        (MONO_INC_FL_ID, None),
        (None, MONO_DEC_FL_ID),
        (MONO_INC_FL_ID, enums.MONOTONICITY_FEATURELIST_DEFAULT),
        (enums.MONOTONICITY_FEATURELIST_DEFAULT, MONO_DEC_FL_ID),
    ],
)
@responses.activate
def test_train_datetime_model_monotonicity(datetime_model, project_id, model_data, featurelists):
    responses.add(
        responses.POST,
        "https://host_name.com/projects/{}/datetimeModels/".format(project_id),
        adding_headers={
            "Location": "https://host_name.com/projects/{}/modelJobs/1/".format(project_id)
        },
        body="",
        status=202,
    )
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/modelJobs/1/".format(project_id),
        body=json.dumps(
            {
                "blueprintId": datetime_model.blueprint_id,
                "featurelistId": datetime_model.featurelist_id,
                "id": 1,
                "isBlocked": False,
                "status": enums.QUEUE_STATUS.COMPLETED,
                "projectId": project_id,
                "modelType": datetime_model.model_type,
                "processes": datetime_model.processes,
            }
        ),
        status=200,
    )

    monotonic_increasing_featurelist_id, monotonic_decreasing_featurelist_id = featurelists

    datetime_model.train_datetime(
        monotonic_increasing_featurelist_id=monotonic_increasing_featurelist_id,
        monotonic_decreasing_featurelist_id=monotonic_decreasing_featurelist_id,
    )

    payload = request_body_to_json(responses.calls[0].request)

    if monotonic_increasing_featurelist_id is enums.MONOTONICITY_FEATURELIST_DEFAULT:
        assert "monotonicIncreasingFeaturelistId" not in payload
    else:
        assert payload["monotonicIncreasingFeaturelistId"] == monotonic_increasing_featurelist_id

    if monotonic_decreasing_featurelist_id is enums.MONOTONICITY_FEATURELIST_DEFAULT:
        assert "monotonicDecreasingFeaturelistId" not in payload
    else:
        assert payload["monotonicDecreasingFeaturelistId"] == monotonic_decreasing_featurelist_id


@responses.activate
def test_get_calculated_prediction_intervals(datetime_model, project_id):
    pred_int_url = "https://host_name.com/projects/{}/models/{}/predictionIntervals/".format(
        project_id, datetime_model.id
    )
    intervals = [80]
    responses.add(
        responses.GET,
        pred_int_url,
        body=json.dumps({"next": None, "data": intervals, "count": 1, "previous": None}),
        status=200,
    )
    retrieved_intervals = datetime_model.get_calculated_prediction_intervals()
    assert intervals == retrieved_intervals


@responses.activate
def test_calculate_prediction_intervals(datetime_model, project_id):
    pred_int_url = "https://host_name.com/projects/{}/models/{}/predictionIntervals/".format(
        project_id, datetime_model.id
    )
    job_url = "https://host_name.com/projects/{}/jobs/1/".format(project_id)

    responses.add(responses.POST, pred_int_url, status=202, headers={"Location": job_url})
    responses.add(
        responses.GET,
        job_url,
        status=303,
        content_type="application/json",
        body=json.dumps(
            {
                "status": u"COMPLETED",
                "model_id": datetime_model.id,
                "is_blocked": False,
                "url": job_url,
                "job_type": "calculate_prediction_intervals",
                "project_id": project_id,
                "id": 1,
            }
        ),
        headers={"Location": "{}result/".format(job_url)},
    )

    job = datetime_model.calculate_prediction_intervals(80)
    assert job.job_type == "calculate_prediction_intervals"
    assert request_body_to_json(responses.calls[0].request) == {"percentiles": [80]}


@responses.activate
def test_compute_datetime_trend_plots(unittest_endpoint, datetime_model, project_id):
    compute_url = "{}/projects/{}/datetimeModels/{}/datetimeTrendPlots/".format(
        unittest_endpoint, project_id, datetime_model.id
    )
    job_url = "{}/projects/{}/jobs/1/".format(unittest_endpoint, project_id)

    responses.add(responses.POST, compute_url, status=202, headers={"Location": job_url})
    responses.add(
        responses.GET,
        job_url,
        status=303,
        content_type="application/json",
        body=json.dumps(
            {
                "status": "COMPLETED",
                "model_id": datetime_model.id,
                "is_blocked": False,
                "url": job_url,
                "job_type": "datetime_trend_plots",
                "project_id": project_id,
                "id": 1,
            }
        ),
        headers={"Location": "{}result/".format(job_url)},
    )

    job = datetime_model.compute_datetime_trend_plots(
        backtest=1,
        source=enums.SOURCE_TYPE.TRAINING,
        forecast_distance_start=1,
        forecast_distance_end=5,
    )
    assert job.job_type == "datetime_trend_plots"
    assert request_body_to_json(responses.calls[0].request) == {
        "backtest": 1,
        "source": enums.SOURCE_TYPE.TRAINING,
        "forecastDistanceStart": 1,
        "forecastDistanceEnd": 5,
    }


@responses.activate
def test_retrieved_model_is_of_datetime_model_from_model_instance(
    datetime_project_server_data, datetime_start_end_model_server_data, project_url,
):
    project_id = datetime_start_end_model_server_data["projectId"]
    model_id = datetime_start_end_model_server_data["id"]

    responses.add(
        responses.GET, project_url, status=200, json=datetime_project_server_data,
    )
    responses.add(
        responses.GET,
        "{}datetimeModels/{}/".format(project_url, model_id),
        status=200,
        json=datetime_start_end_model_server_data,
    )

    model = Model.get(project=project_id, model_id=model_id)
    assert isinstance(model, DatetimeModel)
    assert model.effective_feature_derivation_window_start == -120
    assert model.effective_feature_derivation_window_end == 0



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import datetime

import pytest
import pytz
import requests
import responses
from six.moves.urllib_parse import parse_qs, urlparse
import trafaret

from datarobot import (
    BacktestSpecification,
    DatetimePartitioning,
    DatetimePartitioningSpecification,
    FeatureSettings,
    partitioning_methods,
    Project,
)
from datarobot.helpers.partitioning_methods import Backtest
from datarobot.utils import from_api, parse_time
from tests.test_project import prep_successful_aim_responses
from tests.utils import request_body_to_json


def test_future_proof(datetime_partition_server_data):
    future_response = dict(datetime_partition_server_data, future="newKey")
    DatetimePartitioning.from_server_data(future_response)


def test_construct_duration_string():
    duration = partitioning_methods.construct_duration_string(
        years=1, months=2, days=3, hours=4, minutes=5, seconds=6
    )
    assert duration == "P1Y2M3DT4H5M6S"


def test_construct_duration_string_empty():
    assert partitioning_methods.construct_duration_string() == "P0Y0M0DT0H0M0S"


def test_collect_payload_fails_cleanly_on_type_error(datetime_partition_spec):
    datetime_partition_spec.holdout_start_date = str(datetime_partition_spec.holdout_start_date)
    with pytest.raises(ValueError) as exc_info:
        datetime_partition_spec.collect_payload()
    assert "expected holdout_start_date to be a datetime.datetime" in str(exc_info.value)


def test_backtest_to_specification(datetime_partition):
    backtest = datetime_partition.backtests[0]

    duration_bt_spec = backtest.to_specification()
    assert duration_bt_spec.index == backtest.index
    assert duration_bt_spec.gap_duration == backtest.gap_duration
    assert duration_bt_spec.validation_start_date
    assert duration_bt_spec.validation_duration == backtest.validation_duration

    start_end_bt_spec = backtest.to_specification(use_start_end_format=True)
    assert start_end_bt_spec.index == backtest.index
    assert start_end_bt_spec.primary_training_start_date == backtest.primary_training_start_date
    assert start_end_bt_spec.primary_training_end_date == backtest.primary_training_end_date
    assert start_end_bt_spec.validation_start_date == backtest.validation_start_date
    assert start_end_bt_spec.validation_end_date == backtest.validation_end_date


def test_partition_to_specification(datetime_partition):
    part_spec = datetime_partition.to_specification()
    assert part_spec.datetime_partition_column == datetime_partition.datetime_partition_column
    assert (
        part_spec.autopilot_data_selection_method
        == datetime_partition.autopilot_data_selection_method
    )
    assert part_spec.validation_duration == datetime_partition.validation_duration
    assert part_spec.holdout_start_date == datetime_partition.holdout_start_date
    assert part_spec.holdout_duration == datetime_partition.holdout_duration
    assert part_spec.holdout_end_date is None
    assert part_spec.gap_duration == datetime_partition.gap_duration
    assert part_spec.number_of_backtests == datetime_partition.number_of_backtests
    assert len(part_spec.backtests) == len(datetime_partition.backtests)
    for bt_spec in part_spec.backtests:
        assert isinstance(bt_spec, BacktestSpecification)
        assert bt_spec.primary_training_start_date is None
        assert bt_spec.primary_training_end_date is None
        assert bt_spec.validation_end_date is None
    assert part_spec.use_time_series == datetime_partition.use_time_series
    assert part_spec.default_to_known_in_advance == datetime_partition.default_to_known_in_advance
    assert (
        part_spec.feature_derivation_window_start
        == datetime_partition.feature_derivation_window_start
    )
    assert (
        part_spec.feature_derivation_window_end == datetime_partition.feature_derivation_window_end
    )
    assert part_spec.forecast_window_start == datetime_partition.forecast_window_start
    assert part_spec.forecast_window_end == datetime_partition.forecast_window_end
    assert part_spec.model_splits == datetime_partition.model_splits

    s_e_part_spec = datetime_partition.to_specification(
        use_holdout_start_end_format=True, use_backtest_start_end_format=True
    )
    assert s_e_part_spec.datetime_partition_column == datetime_partition.datetime_partition_column
    assert (
        s_e_part_spec.autopilot_data_selection_method
        == datetime_partition.autopilot_data_selection_method
    )
    assert s_e_part_spec.validation_duration == datetime_partition.validation_duration
    assert s_e_part_spec.holdout_start_date == datetime_partition.holdout_start_date
    assert s_e_part_spec.holdout_end_date == datetime_partition.holdout_end_date
    assert s_e_part_spec.holdout_duration is None
    assert s_e_part_spec.disable_holdout == datetime_partition.disable_holdout
    assert s_e_part_spec.gap_duration == datetime_partition.gap_duration
    assert s_e_part_spec.number_of_backtests == datetime_partition.number_of_backtests
    assert len(s_e_part_spec.backtests) == len(datetime_partition.backtests)
    for bt_spec in s_e_part_spec.backtests:
        assert isinstance(bt_spec, BacktestSpecification)
        assert bt_spec.gap_duration is None
        assert bt_spec.validation_duration is None
    assert s_e_part_spec.use_time_series == datetime_partition.use_time_series
    assert (
        s_e_part_spec.default_to_known_in_advance == datetime_partition.default_to_known_in_advance
    )
    assert (
        s_e_part_spec.feature_derivation_window_start
        == datetime_partition.feature_derivation_window_start
    )
    assert (
        s_e_part_spec.feature_derivation_window_end
        == datetime_partition.feature_derivation_window_end
    )
    assert s_e_part_spec.forecast_window_start == datetime_partition.forecast_window_start
    assert s_e_part_spec.forecast_window_end == datetime_partition.forecast_window_end
    assert s_e_part_spec.model_splits == datetime_partition.model_splits
    assert (
        s_e_part_spec.allow_partial_history_time_series_predictions
        == datetime_partition.allow_partial_history_time_series_predictions
    )


def test_partition_to_specification_with_time_series(datetime_partition):
    datetime_partition.multiseries_id_columns = ["series_id"]
    datetime_partition.differencing_method = "seasonal"
    datetime_partition.treat_as_exponential = "never"
    datetime_partition.periodicities = []
    datetime_partition.use_time_series = True
    datetime_partition.feature_settings = [
        FeatureSettings("special", known_in_advance=True),
        FeatureSettings("special", do_not_derive=True),
    ]
    part_spec = datetime_partition.to_specification()

    assert part_spec.multiseries_id_columns == datetime_partition.multiseries_id_columns
    assert part_spec.differencing_method == datetime_partition.differencing_method
    assert part_spec.treat_as_exponential == datetime_partition.treat_as_exponential
    assert part_spec.periodicities == datetime_partition.periodicities
    assert part_spec.use_time_series == datetime_partition.use_time_series
    assert part_spec.feature_settings == datetime_partition.feature_settings


@responses.activate
def test_retrieve(project_id, project_url, datetime_partition_after_target_server_data):
    responses.add(
        responses.GET,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_after_target_server_data,
    )

    partition = DatetimePartitioning.get(project_id)

    assert partition.primary_training_row_count == 100
    assert partition.available_training_row_count == 100
    assert partition.gap_row_count == 100
    assert partition.holdout_row_count == 100
    assert partition.total_row_count == 100
    assert partition.number_of_known_in_advance_features == 1
    assert partition.number_of_do_not_derive_features == 1
    assert partition.calendar_id == "some-id"
    assert partition.calendar_name == "some name"
    assert partition.model_splits == 2
    assert not partition.allow_partial_history_time_series_predictions

    backtest = partition.backtests[0]
    assert backtest.primary_training_row_count == 100
    assert backtest.available_training_row_count == 100
    assert backtest.gap_row_count == 100
    assert backtest.validation_row_count == 100
    assert backtest.total_row_count == 100


@responses.activate
@pytest.mark.usefixtures("client")
def test_retrieve_time_series(project_id, project_url, datetime_partition_time_series_server_data):
    responses.add(
        responses.GET,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_time_series_server_data,
    )

    partition = DatetimePartitioning.get(project_id)

    fdw_start = datetime_partition_time_series_server_data["featureDerivationWindowStart"]
    fdw_end = datetime_partition_time_series_server_data["featureDerivationWindowEnd"]
    fw_start = datetime_partition_time_series_server_data["forecastWindowStart"]
    fw_end = datetime_partition_time_series_server_data["forecastWindowEnd"]
    assert partition.use_time_series
    assert partition.default_to_known_in_advance
    assert partition.feature_derivation_window_start == fdw_start
    assert partition.feature_derivation_window_end == fdw_end
    assert partition.forecast_window_start == fw_start
    assert partition.forecast_window_end == fw_end
    assert partition.number_of_known_in_advance_features == 1
    assert partition.number_of_do_not_derive_features == 1


@responses.activate
def test_retrieve_time_series_sequence_based_windows(
    project_id, project_url, datetime_partition_time_series_seq_based_windows_server_data
):
    responses.add(
        responses.GET,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_time_series_seq_based_windows_server_data,
    )

    partition = DatetimePartitioning.get(project_id)
    assert partition.use_time_series
    assert partition.windows_basis_unit == "ROW"
    assert partition.periodicities[0].time_unit == "ROW"


@responses.activate
def test_retrieve_time_series_milliseconds(
    project_id, project_url, datetime_partition_milliseconds_server_data
):
    responses.add(
        responses.GET,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_milliseconds_server_data,
    )

    partition = DatetimePartitioning.get(project_id)
    assert partition.use_time_series
    assert partition.windows_basis_unit == "MILLISECOND"
    assert partition.periodicities[0].time_unit == "MILLISECOND"


@responses.activate
def test_generate(
    project_id,
    project_url,
    datetime_partition_server_data,
    holdout_start_date,
    datetime_partition_spec,
):
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_server_data,
    )

    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec)

    # test that we send the right things
    actual_payload = request_body_to_json(responses.calls[0].request)

    expected_backtests = [
        {
            "index": bt.index,
            "gapDuration": bt.gap_duration,
            "validationStartDate": bt.validation_start_date.isoformat(),
            "validationDuration": bt.validation_duration,
        }
        for bt in datetime_partition_spec.backtests
    ]
    expected_payload = {
        "datetimePartitionColumn": datetime_partition_spec.datetime_partition_column,
        "autopilotDataSelectionMethod": datetime_partition_spec.autopilot_data_selection_method,
        "validationDuration": datetime_partition_spec.validation_duration,
        "holdoutStartDate": holdout_start_date.isoformat(),
        "holdoutDuration": datetime_partition_spec.holdout_duration,
        "gapDuration": datetime_partition_spec.gap_duration,
        "numberOfBacktests": datetime_partition_spec.number_of_backtests,
        "backtests": expected_backtests,
        "useTimeSeries": datetime_partition_spec.use_time_series,
        "defaultToKnownInAdvance": datetime_partition_spec.default_to_known_in_advance,
        "defaultToDoNotDerive": datetime_partition_spec.default_to_do_not_derive,
        "modelSplits": datetime_partition_spec.model_splits,
        "allowPartialHistoryTimeSeriesPredictions": (
            datetime_partition_spec.allow_partial_history_time_series_predictions
        ),
    }
    assert actual_payload == expected_payload

    # test that we get the right thing back
    assert partition.project_id == project_id
    assert partition.multiseries_id_columns is None
    assert (
        partition.datetime_partition_column
        == datetime_partition_server_data["datetimePartitionColumn"]
    )
    assert partition.use_time_series == datetime_partition_server_data["useTimeSeries"]
    assert (
        partition.default_to_known_in_advance
        == datetime_partition_server_data["defaultToKnownInAdvance"]
    )
    assert partition.date_format == datetime_partition_server_data["dateFormat"]
    assert (
        partition.autopilot_data_selection_method
        == datetime_partition_server_data["autopilotDataSelectionMethod"]
    )
    assert partition.validation_duration == datetime_partition_server_data["validationDuration"]

    assert isinstance(partition.available_training_start_date, datetime.datetime)
    assert (
        partition.available_training_duration
        == datetime_partition_server_data["availableTrainingDuration"]
    )
    assert isinstance(partition.available_training_end_date, datetime.datetime)

    assert isinstance(partition.primary_training_start_date, datetime.datetime)
    assert (
        partition.primary_training_duration
        == datetime_partition_server_data["primaryTrainingDuration"]
    )
    assert isinstance(partition.primary_training_end_date, datetime.datetime)

    assert isinstance(partition.gap_start_date, datetime.datetime)
    assert partition.gap_duration == datetime_partition_server_data["gapDuration"]
    assert isinstance(partition.gap_end_date, datetime.datetime)

    assert partition.holdout_start_date == holdout_start_date
    assert partition.holdout_duration == datetime_partition_server_data["holdoutDuration"]
    assert isinstance(partition.holdout_end_date, datetime.datetime)

    assert partition.number_of_known_in_advance_features == 1
    assert partition.number_of_do_not_derive_features == 1
    assert partition.model_splits == datetime_partition_server_data["modelSplits"]

    assert partition.number_of_backtests == datetime_partition_server_data["numberOfBacktests"]
    [backtest] = partition.backtests
    [backtest_data] = datetime_partition_server_data["backtests"]
    assert backtest.index == backtest_data["index"]
    assert isinstance(backtest.available_training_start_date, datetime.datetime)
    assert backtest.available_training_duration == backtest_data["availableTrainingDuration"]
    assert isinstance(backtest.available_training_end_date, datetime.datetime)
    assert isinstance(backtest.primary_training_start_date, datetime.datetime)
    assert backtest.primary_training_duration == backtest_data["primaryTrainingDuration"]
    assert isinstance(backtest.primary_training_end_date, datetime.datetime)
    assert isinstance(backtest.gap_start_date, datetime.datetime)
    assert backtest.gap_duration == backtest_data["gapDuration"]
    assert isinstance(backtest.gap_end_date, datetime.datetime)
    assert isinstance(backtest.validation_start_date, datetime.datetime)
    assert backtest.validation_duration == backtest_data["validationDuration"]
    assert isinstance(backtest.validation_end_date, datetime.datetime)


@responses.activate
def test_generate_start_end(
    project_id,
    project_url,
    datetime_partition_server_data,
    datetime_partition_spec,
    holdout_start_date,
    holdout_end_date,
):
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_server_data,
    )

    # replace with start-end fields
    datetime_partition_spec.holdout_duration = None
    datetime_partition_spec.holdout_end_date = parse_time(
        datetime_partition_server_data["holdoutEndDate"]
    )
    datetime_partition_spec.backtests = [
        Backtest.from_data(
            from_api(datetime_partition_server_data["backtests"][0])
        ).to_specification(use_start_end_format=True)
    ]
    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec)

    # test that we send the right things
    actual_payload = request_body_to_json(responses.calls[0].request)

    expected_backtests = [
        {
            "index": bt.index,
            "validationStartDate": bt.validation_start_date.isoformat(),
            "validationEndDate": bt.validation_end_date.isoformat(),
            "primaryTrainingStartDate": bt.primary_training_start_date.isoformat(),
            "primaryTrainingEndDate": bt.primary_training_end_date.isoformat(),
        }
        for bt in datetime_partition_spec.backtests
    ]
    expected_payload = {
        "datetimePartitionColumn": datetime_partition_spec.datetime_partition_column,
        "autopilotDataSelectionMethod": datetime_partition_spec.autopilot_data_selection_method,
        "validationDuration": datetime_partition_spec.validation_duration,
        "holdoutStartDate": holdout_start_date.isoformat(),
        "holdoutEndDate": holdout_end_date.isoformat(),
        "gapDuration": datetime_partition_spec.gap_duration,
        "numberOfBacktests": datetime_partition_spec.number_of_backtests,
        "backtests": expected_backtests,
        "useTimeSeries": datetime_partition_spec.use_time_series,
        "defaultToKnownInAdvance": datetime_partition_spec.default_to_known_in_advance,
        "defaultToDoNotDerive": datetime_partition_spec.default_to_do_not_derive,
        "modelSplits": datetime_partition_spec.model_splits,
        "allowPartialHistoryTimeSeriesPredictions": (
            datetime_partition_spec.allow_partial_history_time_series_predictions
        ),
    }
    assert actual_payload == expected_payload

    # test that we get the right thing back
    assert partition.project_id == project_id
    assert partition.multiseries_id_columns is None
    assert (
        partition.datetime_partition_column
        == datetime_partition_server_data["datetimePartitionColumn"]
    )
    assert partition.use_time_series == datetime_partition_server_data["useTimeSeries"]
    assert (
        partition.default_to_known_in_advance
        == datetime_partition_server_data["defaultToKnownInAdvance"]
    )
    assert partition.date_format == datetime_partition_server_data["dateFormat"]
    assert (
        partition.autopilot_data_selection_method
        == datetime_partition_server_data["autopilotDataSelectionMethod"]
    )
    assert partition.validation_duration == datetime_partition_server_data["validationDuration"]

    assert isinstance(partition.available_training_start_date, datetime.datetime)
    assert (
        partition.available_training_duration
        == datetime_partition_server_data["availableTrainingDuration"]
    )
    assert isinstance(partition.available_training_end_date, datetime.datetime)

    assert isinstance(partition.primary_training_start_date, datetime.datetime)
    assert (
        partition.primary_training_duration
        == datetime_partition_server_data["primaryTrainingDuration"]
    )
    assert isinstance(partition.primary_training_end_date, datetime.datetime)

    assert isinstance(partition.gap_start_date, datetime.datetime)
    assert partition.gap_duration == datetime_partition_server_data["gapDuration"]
    assert isinstance(partition.gap_end_date, datetime.datetime)

    assert partition.holdout_start_date == holdout_start_date
    assert partition.holdout_duration == datetime_partition_server_data["holdoutDuration"]
    assert partition.holdout_end_date == holdout_end_date

    assert partition.number_of_known_in_advance_features == 1
    assert partition.number_of_do_not_derive_features == 1

    assert partition.number_of_backtests == datetime_partition_server_data["numberOfBacktests"]
    [backtest] = partition.backtests
    [backtest_data] = datetime_partition_server_data["backtests"]
    assert backtest.index == backtest_data["index"]
    assert isinstance(backtest.available_training_start_date, datetime.datetime)
    assert backtest.available_training_duration == backtest_data["availableTrainingDuration"]
    assert isinstance(backtest.available_training_end_date, datetime.datetime)
    assert isinstance(backtest.primary_training_start_date, datetime.datetime)
    assert backtest.primary_training_duration == backtest_data["primaryTrainingDuration"]
    assert isinstance(backtest.primary_training_end_date, datetime.datetime)
    assert isinstance(backtest.gap_start_date, datetime.datetime)
    assert backtest.gap_duration == backtest_data["gapDuration"]
    assert isinstance(backtest.gap_end_date, datetime.datetime)
    assert isinstance(backtest.validation_start_date, datetime.datetime)
    assert backtest.validation_duration == backtest_data["validationDuration"]
    assert isinstance(backtest.validation_end_date, datetime.datetime)


@responses.activate
def test_generate_with_multiseries(
    project_id, project_url, datetime_partition_spec, datetime_partition_server_data
):
    datetime_partition_spec.multiseries_id_columns = ["series_id"]
    datetime_partition_server_data["multiseriesIdColumns"] = ["series_id"]
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_server_data,
    )
    multi_prop_url = "{}features/{}/multiseriesProperties/".format(
        project_url, datetime_partition_spec.datetime_partition_column
    )
    multi_prop_json = {
        "datetimePartitionColumn": datetime_partition_spec.datetime_partition_column,
        "detectedMultiseriesIdColumns": [
            {"multiseriesIdColumns": ["series_id"], "timeStep": 1, "timeUnit": "DAY"}
        ],
    }
    date_feat_json = {
        "projectId": project_id,
        "name": datetime_partition_spec.datetime_partition_column,
        "id": 2,
        "timeSeriesEligible": False,
        "featureType": "date",
        "timeSeriesEligibilityReason": "notUnique",
        "lowInformation": False,
        "uniqueCount": 500,
    }
    responses.add(
        responses.GET,
        "{}features/{}/".format(project_url, datetime_partition_spec.datetime_partition_column),
        json=date_feat_json,
    )
    responses.add(responses.GET, multi_prop_url, json=multi_prop_json)

    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec)
    assert partition.multiseries_id_columns == ["series_id"]


@responses.activate
def test_generate_irregular_time_series(
    project_id,
    project_url,
    datetime_partition_irregular_time_series_server_data,
    irregular_time_series_partition_spec,
    holdout_start_date,
):
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_irregular_time_series_server_data,
    )
    spec = irregular_time_series_partition_spec
    partition = DatetimePartitioning.generate(project_id, spec)

    # test that we send the right things
    actual_payload = request_body_to_json(responses.calls[0].request)
    expected_backtests = [
        {
            "index": bt.index,
            "gapDuration": bt.gap_duration,
            "validationStartDate": bt.validation_start_date.isoformat(),
            "validationDuration": bt.validation_duration,
        }
        for bt in spec.backtests
    ]
    expected_payload = {
        "datetimePartitionColumn": spec.datetime_partition_column,
        "autopilotDataSelectionMethod": spec.autopilot_data_selection_method,
        "holdoutStartDate": holdout_start_date.isoformat(),
        "holdoutDuration": spec.holdout_duration,
        "gapDuration": spec.gap_duration,
        "numberOfBacktests": spec.number_of_backtests,
        "backtests": expected_backtests,
        "useTimeSeries": spec.use_time_series,
        "featureDerivationWindowStart": spec.feature_derivation_window_start,
        "featureDerivationWindowEnd": spec.feature_derivation_window_end,
        "forecastWindowStart": spec.forecast_window_start,
        "forecastWindowEnd": spec.forecast_window_end,
        "defaultToKnownInAdvance": False,
        "defaultToDoNotDerive": False,
        "allowPartialHistoryTimeSeriesPredictions": (
            spec.allow_partial_history_time_series_predictions
        ),
    }
    assert actual_payload == expected_payload
    assert partition.validation_duration is None


@responses.activate
def test_generate_time_series(
    project_id,
    project_url,
    datetime_partition_time_series_server_data,
    time_series_partition_spec,
    holdout_start_date,
):
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_time_series_server_data,
    )

    partition = DatetimePartitioning.generate(project_id, time_series_partition_spec)

    # test that we send the right things
    actual_payload = request_body_to_json(responses.calls[0].request)
    expected_backtests = [
        {
            "index": bt.index,
            "gapDuration": bt.gap_duration,
            "validationStartDate": bt.validation_start_date.isoformat(),
            "validationDuration": bt.validation_duration,
        }
        for bt in time_series_partition_spec.backtests
    ]
    expected_payload = {
        "datetimePartitionColumn": time_series_partition_spec.datetime_partition_column,
        "autopilotDataSelectionMethod": time_series_partition_spec.autopilot_data_selection_method,
        "validationDuration": time_series_partition_spec.validation_duration,
        "holdoutStartDate": holdout_start_date.isoformat(),
        "holdoutDuration": time_series_partition_spec.holdout_duration,
        "gapDuration": time_series_partition_spec.gap_duration,
        "numberOfBacktests": time_series_partition_spec.number_of_backtests,
        "backtests": expected_backtests,
        "useTimeSeries": time_series_partition_spec.use_time_series,
        "defaultToKnownInAdvance": time_series_partition_spec.default_to_known_in_advance,
        "defaultToDoNotDerive": time_series_partition_spec.default_to_do_not_derive,
        "featureDerivationWindowStart": time_series_partition_spec.feature_derivation_window_start,
        "featureDerivationWindowEnd": time_series_partition_spec.feature_derivation_window_end,
        "forecastWindowStart": time_series_partition_spec.forecast_window_start,
        "forecastWindowEnd": time_series_partition_spec.forecast_window_end,
        "allowPartialHistoryTimeSeriesPredictions": (
            time_series_partition_spec.allow_partial_history_time_series_predictions
        ),
    }
    assert actual_payload == expected_payload

    assert partition.number_of_known_in_advance_features == 1
    assert partition.number_of_do_not_derive_features == 1
    assert partition.use_time_series == time_series_partition_spec.use_time_series
    assert (
        partition.default_to_known_in_advance
        == time_series_partition_spec.default_to_known_in_advance
    )
    assert (
        partition.feature_derivation_window_start
        == time_series_partition_spec.feature_derivation_window_start
    )
    assert (
        partition.feature_derivation_window_end
        == time_series_partition_spec.feature_derivation_window_end
    )
    assert partition.forecast_window_start == time_series_partition_spec.forecast_window_start
    assert partition.forecast_window_end == time_series_partition_spec.forecast_window_end


@responses.activate
def test_generate_feature_settings_default_values_are_none(
    project_id, project_url, datetime_partition_server_data, datetime_partition_spec
):
    # Here I specifically tests that feature settings which doesn't have any value set for some
    # setting are actually get None instead False for this feature
    feature_settings_data = [
        {"featureName": "input1", "doNotDerive": True},
        {"featureName": "input2", "doNotDerive": False},
        {"featureName": "input3", "knownInAdvance": True},
        {"featureName": "input4", "knownInAdvance": False},
    ]

    feature_settings_specification = [
        FeatureSettings("input1", do_not_derive=True),
        FeatureSettings("input2", do_not_derive=False),
        FeatureSettings("input3", known_in_advance=True),
        FeatureSettings("input4", known_in_advance=False),
    ]

    datetime_partition_server_data["featureSettings"] = feature_settings_data
    datetime_partition_spec.feature_settings = feature_settings_specification

    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_server_data,
    )

    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec)

    # test that we send the right things
    actual_payload = request_body_to_json(responses.calls[0].request)

    expected_feature_settings = [
        {
            "featureName": fs.feature_name,
            "knownInAdvance": fs.known_in_advance,
            "doNotDerive": fs.do_not_derive,
        }
        for fs in datetime_partition_spec.feature_settings
    ]

    # None values get removed by default because we remove None values from request by default
    for setting in expected_feature_settings:
        for key in {"knownInAdvance", "doNotDerive"}:
            if setting[key] is None:
                del setting[key]

    assert actual_payload["featureSettings"] == expected_feature_settings
    assert partition.feature_settings == feature_settings_specification


@responses.activate
def test_generate_feature_settings(
    project_id,
    project_url,
    datetime_partition_server_data_feature_settings,
    holdout_start_date,
    datetime_partition_spec_feature_settings,
):
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_server_data_feature_settings,
    )

    partition = DatetimePartitioning.generate(project_id, datetime_partition_spec_feature_settings)

    # test that we send the right things
    actual_payload = request_body_to_json(responses.calls[0].request)
    expected_backtests = [
        {
            "index": bt.index,
            "gapDuration": bt.gap_duration,
            "validationStartDate": bt.validation_start_date.isoformat(),
            "validationDuration": bt.validation_duration,
        }
        for bt in datetime_partition_spec_feature_settings.backtests
    ]
    expected_feature_settings = [
        {
            "featureName": fs.feature_name,
            "knownInAdvance": fs.known_in_advance,
            "doNotDerive": fs.do_not_derive,
        }
        for fs in datetime_partition_spec_feature_settings.feature_settings
    ]
    # None values get removed by default because we remove None values from request by default
    for setting in expected_feature_settings:
        for key in {"knownInAdvance", "doNotDerive"}:
            if setting[key] is None:
                del setting[key]
    expected_payload = {
        "datetimePartitionColumn": (
            datetime_partition_spec_feature_settings.datetime_partition_column
        ),
        "autopilotDataSelectionMethod": (
            datetime_partition_spec_feature_settings.autopilot_data_selection_method
        ),
        "validationDuration": datetime_partition_spec_feature_settings.validation_duration,
        "holdoutStartDate": holdout_start_date.isoformat(),
        "holdoutDuration": datetime_partition_spec_feature_settings.holdout_duration,
        "gapDuration": datetime_partition_spec_feature_settings.gap_duration,
        "numberOfBacktests": datetime_partition_spec_feature_settings.number_of_backtests,
        "backtests": expected_backtests,
        "useTimeSeries": datetime_partition_spec_feature_settings.use_time_series,
        "defaultToKnownInAdvance": (
            datetime_partition_spec_feature_settings.default_to_known_in_advance
        ),
        "defaultToDoNotDerive": datetime_partition_spec_feature_settings.default_to_do_not_derive,
        "featureSettings": expected_feature_settings,
        "allowPartialHistoryTimeSeriesPredictions": (
            datetime_partition_spec_feature_settings.allow_partial_history_time_series_predictions
        ),
    }
    assert actual_payload == expected_payload
    assert partition.feature_settings == datetime_partition_spec_feature_settings.feature_settings


@pytest.mark.parametrize("known_in_advance", ["True", 1, {"true": True}, [True]])
def test_create_feature_settings_bad_known_in_advance(known_in_advance):
    with pytest.raises(trafaret.DataError):
        FeatureSettings("feature_name", known_in_advance=known_in_advance)


@pytest.mark.parametrize("do_not_derive", ["True", 1, {"true": True}, [True]])
def test_create_feature_settings_bad_do_not_derive(do_not_derive):
    with pytest.raises(trafaret.DataError):
        FeatureSettings("feature_name", do_not_derive=do_not_derive)


@pytest.mark.parametrize("feature_name", [None, False, 1, {"true": True}, [True]])
def test_create_feature_settings_bad_feature_name(feature_name):
    with pytest.raises(trafaret.DataError):
        FeatureSettings(feature_name, "known_in_advance")


@responses.activate
def test_generate_keep_extra_time_series_controls_spec(
    project_id,
    project_url,
    datetime_partition_server_data_extra_time_series_controls,
    datetime_partition_spec_extra_time_series_controls,
):
    server_data = datetime_partition_server_data_extra_time_series_controls
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    responses.add(responses.POST, "{}datetimePartitioning/".format(project_url), json=server_data)

    partitioning = DatetimePartitioning.generate(project_id, spec_fixture)
    partitioning_spec = partitioning.to_specification()

    assert partitioning_spec.periodicities == spec_fixture.periodicities
    assert partitioning_spec.treat_as_exponential == spec_fixture.treat_as_exponential
    assert partitioning_spec.differencing_method == spec_fixture.differencing_method


@responses.activate
def test_generate_no_peridiocities_on_server_spec(
    project_id,
    project_url,
    datetime_partition_server_data_extra_time_series_controls,
    datetime_partition_spec_extra_time_series_controls,
):
    server_data = datetime_partition_server_data_extra_time_series_controls
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    server_data["periodicities"] = None
    spec_fixture.periodicities = None
    responses.add(responses.POST, "{}datetimePartitioning/".format(project_url), json=server_data)

    partitioning = DatetimePartitioning.generate(project_id, spec_fixture)
    partitioning_spec = partitioning.to_specification()

    assert partitioning_spec.periodicities == spec_fixture.periodicities
    assert partitioning_spec.treat_as_exponential == spec_fixture.treat_as_exponential
    assert partitioning_spec.differencing_method == spec_fixture.differencing_method


@responses.activate
def test_create_project_with_periodicity(
    datetime_partition_server_data_extra_time_series_controls,
    datetime_partition_spec_extra_time_series_controls,
):
    server_data = datetime_partition_server_data_extra_time_series_controls
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    prep_successful_aim_responses()
    Project("p-id").set_target("SalePrice", metric="RMSE", partitioning_method=spec_fixture)

    request_message = request_body_to_json(responses.calls[0].request)
    assert request_message["periodicities"] == server_data["periodicities"]


@responses.activate
def test_create_project_with_treat_as_exponential(
    datetime_partition_spec_extra_time_series_controls,
):
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    prep_successful_aim_responses()
    Project("p-id").set_target("SalePrice", metric="RMSE", partitioning_method=spec_fixture)

    request_message = request_body_to_json(responses.calls[0].request)
    assert request_message["treatAsExponential"] == spec_fixture.treat_as_exponential


@responses.activate
def test_create_project_with_differencing_method(
    datetime_partition_spec_extra_time_series_controls,
):
    spec_fixture = datetime_partition_spec_extra_time_series_controls
    prep_successful_aim_responses()
    Project("p-id").set_target("SalePrice", metric="RMSE", partitioning_method=spec_fixture)

    request_message = request_body_to_json(responses.calls[0].request)
    assert request_message["differencingMethod"] == spec_fixture.differencing_method


@responses.activate
def test_retrieve_otp_without_holdout(
    project_id, project_url, datetime_partition_without_holdout_server_data
):
    responses.add(
        responses.GET,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_without_holdout_server_data,
    )

    partition = DatetimePartitioning.get(project_id)

    assert partition.primary_training_start_date is None
    assert partition.primary_training_end_date is None
    assert partition.primary_training_row_count == 0
    assert partition.primary_training_duration == "P0Y0M0D"

    assert partition.available_training_start_date == datetime.datetime(
        2013, 11, 20, 22, 38, 47, tzinfo=pytz.UTC
    )
    assert partition.available_training_end_date == datetime.datetime(
        2014, 1, 12, 22, 44, 23, tzinfo=pytz.UTC
    )
    assert partition.available_training_duration == "P0Y0M53DT0H5M36S"
    assert partition.available_training_row_count == 4276

    assert partition.gap_start_date is None
    assert partition.gap_end_date is None
    assert partition.gap_row_count == 0
    assert partition.gap_duration == "P0Y0M0D"

    assert partition.holdout_start_date is None
    assert partition.holdout_end_date is None
    assert partition.holdout_row_count == 0
    assert partition.holdout_duration == "P0Y0M0D"

    assert partition.number_of_known_in_advance_features == 1
    assert partition.number_of_do_not_derive_features == 1

    assert partition.total_row_count == 4276

    backtest = partition.backtests[0]
    assert backtest.primary_training_row_count == 4168
    assert backtest.available_training_row_count == 4168
    assert backtest.gap_row_count == 0
    assert backtest.validation_row_count == 108
    assert backtest.total_row_count == 4276


@responses.activate
def test_retrieve_feature_settings(
    project_id, project_url, datetime_partition_after_target_server_data_feature_settings
):

    responses.add(
        responses.GET,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_after_target_server_data_feature_settings,
    )

    partition = DatetimePartitioning.get(project_id)

    fs_list = datetime_partition_after_target_server_data_feature_settings["featureSettings"]
    assert partition.feature_settings == [
        FeatureSettings(fs["featureName"], fs["knownInAdvance"]) for fs in fs_list
    ]


@responses.activate
def test_from_minimum_specification(project_id, project_url, datetime_partition_server_data):
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_server_data,
    )

    datetime_column = datetime_partition_server_data["datetimePartitionColumn"]
    min_spec = DatetimePartitioningSpecification(datetime_column)
    DatetimePartitioning.generate(project_id, min_spec)

    expected_payload = {
        "datetimePartitionColumn": datetime_column,
        "useTimeSeries": False,
        "defaultToKnownInAdvance": False,
        "defaultToDoNotDerive": False,
        "allowPartialHistoryTimeSeriesPredictions": False,
    }
    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == expected_payload


def test_to_dataframe(datetime_partition):
    df = datetime_partition.to_dataframe()
    expected_shape = (4 * (1 + datetime_partition.number_of_backtests), 3)
    assert df.shape == expected_shape


@responses.activate
def test_disabled_holdout(project_id, project_url, datetime_partition_without_holdout_server_data):
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_without_holdout_server_data,
    )

    datetime_column = datetime_partition_without_holdout_server_data["datetimePartitionColumn"]
    spec = DatetimePartitioningSpecification(datetime_column, disable_holdout=True)
    DatetimePartitioning.generate(project_id, spec)

    expected_payload = {
        "datetimePartitionColumn": datetime_column,
        "disableHoldout": True,
        "useTimeSeries": False,
        "defaultToKnownInAdvance": False,
        "defaultToDoNotDerive": False,
        "allowPartialHistoryTimeSeriesPredictions": False,
    }
    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == expected_payload


@responses.activate
def test_dt_partitioning_with_unsup(
    project_id, project_url, datetime_partition_without_holdout_server_data
):
    responses.add(
        responses.POST,
        "{}datetimePartitioning/".format(project_url),
        json=datetime_partition_without_holdout_server_data,
    )

    datetime_column = datetime_partition_without_holdout_server_data["datetimePartitionColumn"]
    spec = DatetimePartitioningSpecification(datetime_column, unsupervised_mode=True)
    DatetimePartitioning.generate(project_id, spec)

    expected_payload = {
        "datetimePartitionColumn": datetime_column,
        "useTimeSeries": False,
        "defaultToKnownInAdvance": False,
        "defaultToDoNotDerive": False,
        "unsupervisedMode": True,
        "allowPartialHistoryTimeSeriesPredictions": False,
    }
    actual_payload = request_body_to_json(responses.calls[0].request)
    assert actual_payload == expected_payload


@responses.activate
def test_time_series_feature_log(project_id, project_url):
    expected_resp = {
        "count": 0,
        "previous": None,
        "totalLogLines": 78,
        "featureLog": "",
        "next": None,
    }
    responses.add(responses.GET, "{}timeSeriesFeatureLog/".format(project_url), json=expected_resp)
    resp = DatetimePartitioning.feature_log_list(project_id, limit=10, offset=20)
    assert resp == expected_resp

    req_url = responses.calls[0].request.url
    assert {"limit": ["10"], "offset": ["20"]} == parse_qs(urlparse(req_url).query)


@responses.activate
def test_time_series_feature_log_retrieve(project_id, project_url):
    log_data = "useful info"
    responses.add(responses.GET, "{}timeSeriesFeatureLog/file/".format(project_url), body=log_data)
    retrieved_data = DatetimePartitioning.feature_log_retrieve(project_id)
    assert retrieved_data == log_data


@responses.activate
def test_time_series_optimized_partitioining(
    project_id, project_url, datetime_partition_spec, datetime_partition_after_target_server_data,
):
    responses.add(
        responses.POST,
        "{}optimizedDatetimePartitionings/".format(project_url),
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": "https://host_name.com/status/status-id/"},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/status/status-id/",
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": "{}optimizedDatetimePartitionings/".format(project_url)},
    )
    responses.add(
        responses.GET,
        "{}optimizedDatetimePartitionings/".format(project_url),
        status=200,
        json=datetime_partition_after_target_server_data,
        content_type="application/json",
    )

    # Assert that the synchronous route response is not implemented
    with pytest.raises(requests.exceptions.ConnectionError):
        DatetimePartitioning.generate(project_id=project_id, spec=datetime_partition_spec)

    # Assert that 'generate_optimized' works correctly
    partition = DatetimePartitioning.generate_optimized(
        project_id=project_id, spec=datetime_partition_spec, target="target"
    )
    assert partition.validation_duration == "P14D"

    # Assert that sending a target to `generate` works the same as using `generate_optimized`
    partition = DatetimePartitioning.generate(
        project_id=project_id, spec=datetime_partition_spec, target="target"
    )
    assert partition.validation_duration == "P14D"



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import pytest
import responses

import datarobot
from datarobot import Blueprint, enums, Project
from tests.utils import request_body_to_json


@responses.activate
def test_async_flow(project, project_id, model_job_json, model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type="application/json",
        status=202,
        body="",
        adding_headers={"Location": "https://host_name.com/projects/p-id/modelJobs/12/"},
    )
    responses.add(
        responses.GET,
        "https://host_name.com/projects/{}/modelJobs/12/".format(project_id),
        content_type="application/json",
        status=200,
        body=model_job_json,
    )

    blueprint_id = "e1c7fc29ba2e612a72272324b8a842af"
    retval = project.train(blueprint_id)
    assert retval == "12"


def test_print_project_from_id():
    project = Project("some-project-id")
    print(project)  # checks that method relevant to print works on this (failed in one release)


@responses.activate
def test_all_defaults(project, model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type="application/json",
        status=201,
        body="",
        adding_headers={"Location": "http://host_name.com/projects/p-id/modelJobs/id/"},
    )
    blueprint_id = "e1c7fc29ba2e612a72272324b8a842af"
    model_id = project.train(blueprint_id)
    assert responses.calls[0].request.url == model_collection_url
    payload = request_body_to_json(responses.calls[0].request)
    assert "samplePct" not in payload
    assert "trainingRowCount" not in payload
    assert "id" == model_id


@responses.activate
def test_no_defaults(project, model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type="application/json",
        status=201,
        body="",
        adding_headers={"Location": "http://host_name.com/projects/p-id/modelJobs/id/"},
    )
    blueprint_id = "e1c7fc29ba2e612a72272324b8a842af"
    dataset_id = "5223deadbeefdeadbeef0101"
    source_project_id = "5223deadbeefdeadbeef1234"
    scoring_type = datarobot.SCORING_TYPE.cross_validation
    sample_pct = 44

    project.train(
        blueprint_id,
        featurelist_id=dataset_id,
        source_project_id=source_project_id,
        sample_pct=sample_pct,
        scoring_type=scoring_type,
    )
    assert responses.calls[0].request.url == model_collection_url
    request = request_body_to_json(responses.calls[0].request)
    assert request["blueprintId"] == blueprint_id
    assert request["sourceProjectId"] == source_project_id
    assert request["featurelistId"] == dataset_id
    assert request["samplePct"] == sample_pct
    assert request["scoringType"] == scoring_type
    assert "trainingRowCount" not in request


def test_train_cant_use_both(project):
    blueprint_id = "a-blueprint"
    with pytest.raises(ValueError):
        project.train(blueprint_id, sample_pct=40, training_row_count=100)


@responses.activate
def test_train_by_rowcount_with_bp_id(project, model_collection_url, project_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type="application/json",
        status=201,
        body="",
        adding_headers={"Location": "{}/modelJobs/5/".format(project_url)},
    )
    blueprint_id = "a-blueprint-id-hash"
    training_row_count = 50
    project.train(blueprint_id, training_row_count=training_row_count)
    assert responses.calls[0].request.url == model_collection_url
    payload = request_body_to_json(responses.calls[0].request)
    assert payload["blueprintId"] == blueprint_id
    assert payload["trainingRowCount"] == training_row_count
    assert "samplePct" not in payload


@responses.activate
def test_train_by_rowcount_with_trainable(project, model_collection_url, project_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type="application/json",
        status=201,
        body="",
        adding_headers={"Location": "{}/modelJobs/5/".format(project_url)},
    )
    blueprint_id = "a-blueprint-hash"
    trainable = Blueprint.from_data(
        {
            "id": blueprint_id,
            "project_id": project.id,
            "model_type": "A fake model",
            "processes": ["Things", "Stuff"],
        }
    )
    training_row_count = 50

    project.train(trainable, training_row_count=training_row_count)
    payload = request_body_to_json(responses.calls[0].request)
    assert payload["blueprintId"] == blueprint_id
    assert payload["trainingRowCount"] == training_row_count
    assert "samplePct" not in payload


@responses.activate
def test_with_trainable_object_ignores_any_source_project_id(project, model_collection_url):

    responses.add(
        responses.POST,
        model_collection_url,
        content_type="application/json",
        status=201,
        body="",
        adding_headers={"Location": "http://pam/api/v2/projects/p-id/models/id/"},
    )
    blueprint_id = "e1c7fc29ba2e612a72272324b8a842af"
    source_project_id = "5223deadbeefdeadbeef1234"

    data = dict(
        id=blueprint_id,
        project_id=source_project_id,
        model_type="Pretend Model",
        processes=["Cowboys", "Aliens"],
    )

    blueprint = Blueprint.from_data(data)

    project.train(blueprint, source_project_id="should-be-ignored")
    assert responses.calls[0].request.url == model_collection_url
    request = request_body_to_json(responses.calls[0].request)
    assert request["blueprintId"] == blueprint_id
    assert request["sourceProjectId"] == source_project_id


MONO_INC_FL_ID = "5ae1afb6962d74625f940991"
MONO_DEC_FL_ID = "5ae1afb6962d74625f940992"
BLUEPRINT_ID = "5a997601962d742269a38dd9"


@pytest.mark.parametrize(
    "blueprint,featurelists",
    [
        (Blueprint(id=BLUEPRINT_ID), (MONO_INC_FL_ID, MONO_DEC_FL_ID)),
        (BLUEPRINT_ID, (MONO_INC_FL_ID, MONO_DEC_FL_ID)),
        (BLUEPRINT_ID, (None, None)),
        (BLUEPRINT_ID, (MONO_INC_FL_ID, None)),
        (BLUEPRINT_ID, (None, MONO_DEC_FL_ID)),
        (BLUEPRINT_ID, (enums.MONOTONICITY_FEATURELIST_DEFAULT, MONO_DEC_FL_ID)),
    ],
)
@responses.activate
def test_monotonicity_blueprint(blueprint, project, featurelists):
    monotonic_increasing_featurelist_id, monotonic_decreasing_featurelist_id = featurelists

    responses.add(
        responses.POST,
        "https://host_name.com/projects/{}/models/".format(project.id),
        content_type="application/json",
        status=201,
        body="",
        adding_headers={
            "Location": "https://host_name.com/projects/p-id/5a9975c9962d746034393e42/modelJobs/1/"
        },
    )

    project.train(
        blueprint,
        monotonic_increasing_featurelist_id=monotonic_increasing_featurelist_id,
        monotonic_decreasing_featurelist_id=monotonic_decreasing_featurelist_id,
    )
    payload = request_body_to_json(responses.calls[0].request)
    if monotonic_increasing_featurelist_id is enums.MONOTONICITY_FEATURELIST_DEFAULT:
        assert "monotonicIncreasingFeaturelistId" not in payload
    else:
        assert payload["monotonicIncreasingFeaturelistId"] == monotonic_increasing_featurelist_id

    if monotonic_decreasing_featurelist_id is enums.MONOTONICITY_FEATURELIST_DEFAULT:
        assert "monotonicDecreasingFeaturelistId" not in payload
    else:
        assert payload["monotonicDecreasingFeaturelistId"] == monotonic_decreasing_featurelist_id



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import dateutil
import pytest
import responses

from datarobot import Deployment
from datarobot.enums import DATA_DRIFT_METRIC
from datarobot.models.deployment import FeatureDrift, TargetDrift


class TestTargetDrift(object):
    @pytest.fixture
    def response_data(self, deployment_data):
        return {
            "model_id": deployment_data["model"]["id"],
            "period": {
                "start": "2020-04-17T19:00:00.000000Z",
                "end": "2020-04-24T19:00:00.000000Z",
            },
            "metric": "psi",
            "target_name": "readmitted",
            "drift_score": 0.0040851491404388825,
            "sample_size": 5333,
            "baseline_sample_size": 504507,
        }

    @pytest.fixture
    def get_response(self, unittest_endpoint, deployment_data, response_data):
        url = "{}/deployments/{}/targetDrift/".format(unittest_endpoint, deployment_data["id"])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(response_data),
        )

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "get_response")
    @pytest.mark.parametrize("from_deployment", [True, False])
    def test_target_drift(self, deployment_data, response_data, from_deployment):
        if from_deployment:
            target_drift = Deployment.get(deployment_data["id"]).get_target_drift()
        else:
            target_drift = TargetDrift.get(deployment_data["id"])

        expected_start_date = dateutil.parser.parse(response_data["period"]["start"])
        expected_end_date = dateutil.parser.parse(response_data["period"]["end"])
        assert target_drift.period["start"] == expected_start_date
        assert target_drift.period["end"] == expected_end_date

        assert target_drift.model_id == response_data["model_id"]
        assert target_drift.metric == response_data["metric"]
        assert target_drift.target_name == response_data["target_name"]
        assert target_drift.drift_score == response_data["drift_score"]
        assert target_drift.sample_size == response_data["sample_size"]
        assert target_drift.baseline_sample_size == response_data["baseline_sample_size"]

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response")
    @pytest.mark.parametrize("metric", DATA_DRIFT_METRIC.ALL)
    def test_metric_query_param(self, deployment_data, response_data, metric, unittest_endpoint):
        response_data["metric"] = metric
        url = "{}/deployments/{}/targetDrift/?metric={}".format(
            unittest_endpoint, deployment_data["id"], metric
        )
        responses.add(
            responses.GET,
            url,
            status=200,
            match_querystring=True,
            content_type="application/json",
            body=json.dumps(response_data),
        )

        target_drift = Deployment.get(deployment_data["id"]).get_target_drift(metric=metric)
        assert target_drift.metric == metric
        target_drift = TargetDrift.get(deployment_data["id"], metric=metric)
        assert target_drift.metric == metric


class TestFeatureDrift(object):
    @pytest.fixture
    def response_data_first_page(self, deployment_data, unittest_endpoint):
        return {
            "model_id": deployment_data["model"]["id"],
            "metric": "psi",
            "period": {
                "start": "2020-04-20T19:00:00.000000Z",
                "end": "2020-04-27T19:00:00.000000Z",
            },
            "count": 2,
            "next": "{}/deployments/{}/featureDrift/?offset=2&limit=2&metric=psi".format(
                unittest_endpoint, deployment_data["id"]
            ),
            "previous": None,
            "data": [
                {
                    "featureImpact": 1.0,
                    "sampleSize": 157,
                    "name": "age",
                    "baselineSampleSize": 2018030,
                    "driftScore": 4.169815947131828,
                },
                {
                    "featureImpact": 0.751510387892,
                    "sampleSize": 120,
                    "name": "diag_1",
                    "baselineSampleSize": 2018030,
                    "driftScore": 1.2364572584578573,
                },
            ],
        }

    @pytest.fixture
    def response_data_second_page(self, deployment_data, unittest_endpoint):
        return {
            "model_id": deployment_data["model"]["id"],
            "metric": "psi",
            "period": {
                "start": "2020-04-20T19:00:00.000000Z",
                "end": "2020-04-27T19:00:00.000000Z",
            },
            "count": 1,
            "next": None,
            "previous": None,
            "data": [
                {
                    "featureImpact": 0.609462873473389,
                    "sampleSize": 147,
                    "name": "gender",
                    "baselineSampleSize": 2018030,
                    "driftScore": 3.3003671706815156,
                },
            ],
        }

    @pytest.fixture
    def get_response(
        self,
        unittest_endpoint,
        deployment_data,
        response_data_first_page,
        response_data_second_page,
    ):
        url = "{}/deployments/{}/featureDrift/".format(unittest_endpoint, deployment_data["id"])
        responses.add(
            responses.GET,
            url,
            status=200,
            content_type="application/json",
            body=json.dumps(response_data_first_page),
        )
        responses.add(
            responses.GET,
            response_data_first_page["next"],
            status=200,
            content_type="application/json",
            body=json.dumps(response_data_second_page),
        )

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response", "get_response")
    @pytest.mark.parametrize("from_deployment", [True, False])
    def test_feature_drift(
        self, deployment_data, response_data_first_page, response_data_second_page, from_deployment,
    ):
        if from_deployment:
            feature_drifts = Deployment.get(deployment_data["id"]).get_feature_drift()
        else:
            feature_drifts = FeatureDrift.list(deployment_data["id"])

        all_features = response_data_first_page["data"] + response_data_second_page["data"]
        expected_start_date = dateutil.parser.parse(response_data_first_page["period"]["start"])
        expected_end_date = dateutil.parser.parse(response_data_first_page["period"]["end"])
        for index, feature_drift in enumerate(feature_drifts):
            assert feature_drift.period["start"] == expected_start_date
            assert feature_drift.period["end"] == expected_end_date

            assert feature_drift.model_id == response_data_first_page["model_id"]
            assert feature_drift.metric == response_data_first_page["metric"]
            assert feature_drift.name == all_features[index]["name"]
            assert feature_drift.drift_score == all_features[index]["driftScore"]
            assert feature_drift.feature_impact == all_features[index]["featureImpact"]
            assert feature_drift.sample_size == all_features[index]["sampleSize"]
            assert feature_drift.baseline_sample_size == all_features[index]["baselineSampleSize"]

    @responses.activate
    @pytest.mark.usefixtures("deployment_get_response")
    @pytest.mark.parametrize("metric", DATA_DRIFT_METRIC.ALL)
    def test_metric_query_param(self, deployment_data, metric, unittest_endpoint):
        response_data = {
            "model_id": deployment_data["model"]["id"],
            "metric": metric,
            "period": {
                "start": "2020-04-20T19:00:00.000000Z",
                "end": "2020-04-27T19:00:00.000000Z",
            },
            "count": 1,
            "next": None,
            "previous": None,
            "data": [
                {
                    "featureImpact": 1.0,
                    "sampleSize": 157,
                    "name": "age",
                    "baselineSampleSize": 2018030,
                    "driftScore": 4.169815947131828,
                },
            ],
        }
        url = "{}/deployments/{}/featureDrift/?metric={}".format(
            unittest_endpoint, deployment_data["id"], metric
        )
        responses.add(
            responses.GET,
            url,
            status=200,
            match_querystring=True,
            content_type="application/json",
            body=json.dumps(response_data),
        )

        feature_drift = Deployment.get(deployment_data["id"]).get_feature_drift(metric=metric)
        for feature in feature_drift:
            assert feature.metric == metric
        feature_drift = FeatureDrift.list(deployment_data["id"], metric=metric)
        for feature in feature_drift:
            assert feature.metric == metric



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import mock
import pytest
import responses


@pytest.yield_fixture
def mock_async_time():
    with mock.patch("datarobot.utils.waiters.time") as mock_time:
        yield mock_time


@pytest.fixture
def async_url():
    return "https://host_name.com/status/status-id/"


@pytest.fixture
def model_collection_url(project_id):
    return "https://host_name.com/projects/{}/models/".format(project_id)


@pytest.fixture
def model_job_data(project_id, model_id):
    return {
        "status": "inprogress",
        "samplePct": 64.0,
        "processes": [
            "One-Hot Encoding",
            "Missing Values Imputed",
            "Standardize",
            "Linear Regression",
        ],
        "modelType": "Linear Regression",
        "featurelistId": "55666d05100d2b01a1104dae",
        "blueprintId": "3bb4665320be633b30a9601b3e73284d",
        "projectId": project_id,
        "id": 12,
        "modelId": model_id,
    }


@pytest.fixture
def model_job_json(model_job_data):
    return json.dumps(model_job_data)


@pytest.fixture
def project_creation_responses(
    async_url, project_collection_url, project_url, project_without_target_json
):
    responses.add(
        responses.POST,
        project_collection_url,
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": async_url},
    )
    responses.add(
        responses.GET,
        async_url,
        status=303,
        body="",
        content_type="application/json",
        adding_headers={"Location": project_url},
    )
    responses.add(
        responses.GET,
        project_url,
        status=200,
        body=project_without_target_json,
        content_type="application/json",
    )


@pytest.fixture
def project_aim_responses(project_aim_url, async_url, project_url, project_with_target_json):
    responses.add(
        responses.PATCH,
        project_aim_url,
        body="",
        status=202,
        adding_headers={"Location": async_url},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        async_url,
        body="",
        status=303,
        adding_headers={"Location": project_url},
        content_type="application/json",
    )

    responses.add(
        responses.GET,
        project_url,
        status=200,
        content_type="application/json",
        body=project_with_target_json,
    )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from conf import version

if __name__ == "__main__":
    print(version)



# coding=utf-8
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

from numpy import isclose
import pytest
import responses

from datarobot import Model
from datarobot.models.word_cloud import WordCloud


@pytest.fixture
def word_cloud_data():
    return {
        "ngrams": [
            {
                "ngram": u"エピソード",
                "coefficient": 0.2074072186419962,
                "isStopword": False,
                "count": 10,
                "frequency": 0.1449275362318841,
            },
            {
                "ngram": u"外",
                "coefficient": 0.3669030452411179,
                "isStopword": True,
                "count": 8,
                "frequency": 0.1159420289855072,
            },
            {
                "ngram": u"閉鎖",
                "coefficient": -0.1343662758815528,
                "isStopword": False,
                "count": 5,
                "frequency": 0.07246376811594203,
            },
        ]
    }


@pytest.fixture
def word_cloud_data_with_class(word_cloud_data):
    for ngram in word_cloud_data["ngrams"]:
        ngram["class"] = "classname"
    return word_cloud_data


@pytest.fixture
def word_cloud_data_with_multiple_classes(word_cloud_data):
    word_cloud_data["ngrams"][0]["class"] = "class_0"
    word_cloud_data["ngrams"][1]["class"] = "class_1"
    word_cloud_data["ngrams"][2]["class"] = "class_2"
    return word_cloud_data


@pytest.fixture
def word_cloud_data_with_variable(word_cloud_data_with_class):
    for ngram in word_cloud_data_with_class["ngrams"]:
        ngram["variable"] = "SOURCE_COLUMN"
    return word_cloud_data_with_class


def test_instantiation(word_cloud_data):
    wc = WordCloud.from_server_data(word_cloud_data)

    assert len(word_cloud_data["ngrams"]) == len(wc.ngrams)
    for i in range(len(wc.ngrams)):
        assert wc.ngrams[i]["ngram"] == word_cloud_data["ngrams"][i]["ngram"]
        assert wc.ngrams[i]["is_stopword"] == word_cloud_data["ngrams"][i]["isStopword"]
        assert wc.ngrams[i]["count"] == word_cloud_data["ngrams"][i]["count"]
        assert isclose(wc.ngrams[i]["coefficient"], word_cloud_data["ngrams"][i]["coefficient"])
        assert isclose(wc.ngrams[i]["frequency"], word_cloud_data["ngrams"][i]["frequency"])
        # Backward compatible - for old return results fill with None
        assert wc.ngrams[i]["class"] is None


def test_instantiation_with_class(word_cloud_data_with_class):
    wc = WordCloud.from_server_data(word_cloud_data_with_class)
    assert len(word_cloud_data_with_class["ngrams"]) == len(wc.ngrams)
    for i in range(len(wc.ngrams)):
        assert wc.ngrams[i]["class"] == "classname"


def test_ngrams_per_class(word_cloud_data_with_multiple_classes):
    wc = WordCloud.from_server_data(word_cloud_data_with_multiple_classes)
    ngrams_per_class = wc.ngrams_per_class()
    assert set(ngrams_per_class.keys()) == {"class_0", "class_1", "class_2"}


def test_instantiation_with_variable(word_cloud_data_with_variable):
    wc = WordCloud.from_server_data(word_cloud_data_with_variable)
    assert len(word_cloud_data_with_variable["ngrams"]) == len(wc.ngrams)
    assert all(ngram["variable"] == "SOURCE_COLUMN" for ngram in wc.ngrams)


def test_future_proof(word_cloud_data):
    data_with_future_keys = dict(word_cloud_data, new_key="new_value")
    data_with_future_keys["ngrams"][0]["some_additional_data"] = "we never expected this"
    WordCloud.from_server_data(data_with_future_keys)


def test_most_frequent(word_cloud_data):
    wc = WordCloud.from_server_data(word_cloud_data)
    most_frequent_list = wc.most_frequent(1)
    assert len(most_frequent_list) == 1
    most_frequent = most_frequent_list[0]
    assert all(most_frequent["frequency"] >= wc_ngram["frequency"] for wc_ngram in wc.ngrams)


def test_most_important(word_cloud_data):
    wc = WordCloud.from_server_data(word_cloud_data)
    most_important_list = wc.most_important(1)
    assert len(most_important_list) == 1
    most_important = most_important_list[0]
    assert all(
        abs(most_important["coefficient"]) >= abs(wc_ngram["coefficient"]) for wc_ngram in wc.ngrams
    )


@pytest.fixture
def word_cloud_url(project_id, model_id):
    return "https://host_name.com/projects/{}/models/{}/wordCloud/".format(project_id, model_id)


@responses.activate
def test_get_word_cloud(word_cloud_data, word_cloud_url, project_id, model_id):
    responses.add(
        responses.GET,
        word_cloud_url,
        status=200,
        content_type="application/json",
        body=json.dumps(word_cloud_data),
    )
    model = Model(id=model_id, project_id=project_id)
    wc = model.get_word_cloud()

    assert len(word_cloud_data["ngrams"]) == len(wc.ngrams)
    for i in range(len(wc.ngrams)):
        assert wc.ngrams[i]["ngram"] == word_cloud_data["ngrams"][i]["ngram"]
        assert wc.ngrams[i]["is_stopword"] == word_cloud_data["ngrams"][i]["isStopword"]
        assert wc.ngrams[i]["count"] == word_cloud_data["ngrams"][i]["count"]
        assert isclose(wc.ngrams[i]["coefficient"], word_cloud_data["ngrams"][i]["coefficient"])
        assert isclose(wc.ngrams[i]["frequency"], word_cloud_data["ngrams"][i]["frequency"])



#! /usr/bin/env python3
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""Show a small sample of duplicate images in a project.

The following will open a project, get the first model id where the
feature name matches, and then display a few of the duplicate images
in the GUI.
"""
import io

import PIL.Image

from datarobot.models import Project
from datarobot.models.visualai import DuplicateImage


def display_images(project_name, max_images):
    project = Project.list(search_params={"project_name": project_name})[0]
    for image in DuplicateImage.list(project.id, "image")[:max_images]:
        with io.BytesIO(image.image_bytes) as bio, PIL.Image.open(bio) as img:
            img.show()


if __name__ == "__main__":
    project_name = "dataset_2k.zip"
    max_images = 2
    display_images(project_name, max_images)



# coding: utf-8
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime

import dateutil
import pytest
import responses

from datarobot import FrozenModel
from tests.utils import assert_equal_py2, assert_equal_py3


@responses.activate
def test_frozen_model_get_returns_valid_data(frozen_json, one_frozen):
    url = "https://host_name.com/projects/p-id/frozenModels/{}/".format(one_frozen.id)
    responses.add(responses.GET, url, body=frozen_json, status=200, content_type="application/json")
    frozen = FrozenModel.get("p-id", one_frozen.id)
    assert isinstance(frozen, FrozenModel)
    assert frozen.id == one_frozen.id
    assert frozen.parent_model_id == one_frozen.parent_model_id
    assert isinstance(frozen.model_number, int)


@responses.activate
@pytest.mark.parametrize(
    "sample_pct,training_row_count", [(None, None), (98.98, None), (None, 1000)]
)
def test_request_frozen_model(
    one_model, project_url, frozen_model_job_completed_server_data, sample_pct, training_row_count
):
    frozen_models_url = "{}frozenModels/".format(project_url)
    job_url = "{}modelJobs/{}/".format(project_url, frozen_model_job_completed_server_data["id"])
    finished_freeze_url = "{}/12344321beefcafe43211234/".format(frozen_models_url)

    responses.add(
        responses.POST,
        frozen_models_url,
        body="",
        status=202,
        content_type="application/json",
        adding_headers={"Location": job_url},
    )
    responses.add(
        responses.GET,
        job_url,
        json=frozen_model_job_completed_server_data,
        status=303,
        adding_headers={"Location": finished_freeze_url},
    )
    if sample_pct:
        model_job = one_model.request_frozen_model(sample_pct=sample_pct)
    elif training_row_count:
        model_job = one_model.request_frozen_model(training_row_count=training_row_count)
    else:
        model_job = one_model.request_frozen_model()
    assert model_job.id == int(frozen_model_job_completed_server_data["id"])


@pytest.fixture
def datetime_frozen_data(datetime_model_data, frozen_data):
    modified_data = dict(frozen_data)
    modified_data.update(datetime_model_data)
    modified_data.pop("sample_pct")
    return modified_data


def test_datetime_frozen_model(datetime_frozen_data):
    mod = FrozenModel.from_data(datetime_frozen_data)
    dt = mod.training_start_date
    assert dt.tzinfo == dateutil.tz.tzutc()
    assert isinstance(mod.training_end_date, datetime)
    assert mod.training_row_count is None
    assert mod.training_duration is None


class TestRepr(object):
    """Tests of method FrozenModel.__repr__"""

    def test_model_type_str(self):
        # when both model_type and id are set
        frozen_model = FrozenModel(id="id", model_type="model_type")
        # then repr contains only model_type
        assert repr(frozen_model) == "FrozenModel('model_type')"

    def test_model_type_unicode(self):
        # when model_type contains non-ascii characters
        frozen_model = FrozenModel(model_type=u"тайп")
        # then repr is an ascii string
        assert_equal_py2(repr(frozen_model), "FrozenModel(u'\\u0442\\u0430\\u0439\\u043f')")
        assert_equal_py3(repr(frozen_model), "FrozenModel('тайп')")

    def test_id(self):
        # when id is set and model_type is missing
        frozen_model = FrozenModel(id="58668cb5bf36cd77a5d8056c")
        # then repr contains id
        assert repr(frozen_model) == "FrozenModel('58668cb5bf36cd77a5d8056c')"



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import csv
import os

import pandas as pd
import six

from .. import errors


def dataframe_to_buffer(df):
    """Convert a dataframe to a serialized form in a buffer

    Parameters
    ----------
    df : pandas.DataFrame
        The data to serialize

    Returns
    -------
    buff : StringIO()
        The data. The descriptor will be reset before being returned (seek(0))
    """
    buff = six.StringIO()
    df.to_csv(buff, encoding="utf-8", index=False, quoting=csv.QUOTE_ALL)
    buff.seek(0)
    return buff


def list_of_records_to_buffer(list_of_records):
    """

    Parameters
    ----------
    list_of_records: list(dict)
        A list of records to serialize. Each row should be a dict with the same keys as the first
        row.

    Returns
    -------
    buffer: six.StringIO
        The data. The descriptor will be reset before being returned (seek(0))
    """
    buff = six.StringIO()
    headers = [key for key in list_of_records[0]]
    csv_writer = csv.DictWriter(buff, fieldnames=headers)
    csv_writer.writeheader()
    csv_writer.writerows(list_of_records)
    buff.seek(0)
    return buff


def is_urlsource(sourcedata):
    """ Whether sourcedata is of url kind
    """
    return isinstance(sourcedata, six.string_types) and (
        sourcedata.startswith("http")
        or sourcedata.startswith("file:")
        or sourcedata.startswith("s3:")
    )


def recognize_sourcedata(sourcedata, default_fname):
    """Given a sourcedata figure out if it is a filepath, dataframe, or
    filehandle, and then return the correct kwargs for the upload process
    """
    if isinstance(sourcedata, pd.DataFrame):
        buff = dataframe_to_buffer(sourcedata)
        return {"filelike": buff, "fname": default_fname}
    elif hasattr(sourcedata, "read") and hasattr(sourcedata, "seek"):
        return {"filelike": sourcedata, "fname": default_fname}
    elif isinstance(sourcedata, six.string_types) and os.path.isfile(sourcedata):
        return {"file_path": sourcedata, "fname": os.path.split(sourcedata)[1]}
    elif isinstance(sourcedata, six.binary_type) and not is_urlsource(sourcedata):
        assert_modelable(sourcedata)
        return {"content": sourcedata, "fname": default_fname}
    else:
        # Checking if string representation of an object can be used as a file path.
        # The main reason for this is supporting Python's pathlib.Path without having extra
        # dependency for python lower than 3.4.
        try:
            if os.path.isfile(str(sourcedata)):
                return {"file_path": str(sourcedata), "fname": os.path.split(str(sourcedata))[1]}
        except Exception:
            pass

    err_msg = (
        "sourcedata parameter not understood. Use pandas "
        "DataFrame, file object or string that is either a "
        "path to file or raw file content to specify data "
        "to upload"
    )
    raise errors.InputNotUnderstoodError(err_msg)


def assert_modelable(sourcedata):
    """
    Uses a heuristic to assert that the given argument is not a
    filepath.

    Some users have mistyped filepaths before, which
    the function `recognize_sourcedata` interpreted as being
    the actual data to use for modeling. This would lead to other
    problems later on.

    Parameters
    ----------
    sourcedata : six.binary_type
        The data which we are trying to assert is not a mistyped
        file path.

    Raises
    ------
    InputNotUnderstoodError
        If this does look like a filepath (it should not)
    """
    if len(sourcedata) < 500 and len(sourcedata.splitlines()) == 1:
        raise errors.InputNotUnderstoodError(
            "The source appears to contain only a single line of data. "
            "Did you mistype a file path?"
        )
